# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

"""
Metric classes for Ax benchmarking.

Metrics vary on two dimensions: Whether they are `MapMetric`s or not, and
whether they are available while running or not.

There are four Metric classes:
- `BenchmarkMetric`: A non-Map metric
    is not available while running.
- `BenchmarkMapMetric`: For when outputs should be `MapData` (not `Data`) and
    data is available while running.
- `BenchmarkTimeVaryingMetric`: For when outputs should be `Data` and the metric
  is available while running.
- `BenchmarkMapUnavailableWhileRunningMetric`: For when outputs should be
  `MapData` and the metric is not available while running.

Any of these can be used with or without a simulator. However,
`BenchmarkMetric.fetch_trial_data` cannot take in data with multiple time steps,
as they will not be used and this is assumed to be an error. The below table
enumerates use cases.

.. list-table:: Benchmark Metrics Table
   :widths: 5 25 5 5 5 50
   :header-rows: 1

   * -
     - Metric
     - Map
     - Available while running
     - Simulator
     - Reason/use case
   * - 1
     - BenchmarkMetric
     - No
     - No
     - No
     - Vanilla
   * - 2
     - BenchmarkMetric
     - No
     - No
     - Yes
     - Asynchronous, data read only at end
   * - 3
     - BenchmarkTimeVaryingMetric
     - No
     - Yes
     - No
     - Behaves like #1 because it will never be RUNNING
   * - 4
     - BenchmarkTimeVaryingMetric
     - No
     - Yes
     - Yes
     - Scalar data that changes over time
   * - 5
     - BenchmarkMapUnavailableWhileRunningMetric
     - Yes
     - No
     - No
     - MapData that returns immediately; could be used for getting baseline
   * - 6
     - BenchmarkMapUnavailableWhileRunningMetric
     - Yes
     - No
     - Yes
     - Asynchronicity with MapData read only at end
   * - 7
     - BenchmarkMapMetric
     - Yes
     - Yes
     - No
     - Behaves same as #5
   * - 8
     - BenchmarkMapMetric
     - Yes
     - Yes
     - Yes
     - Early stopping
"""

from abc import abstractmethod
from typing import Any

import numpy as np
from ax.benchmark.benchmark_trial_metadata import BenchmarkTrialMetadata
from ax.core.base_trial import BaseTrial
from ax.core.batch_trial import BatchTrial
from ax.core.data import Data
from ax.core.map_data import MapData
from ax.core.map_metric import MapMetric
from ax.core.metric import Metric, MetricFetchE, MetricFetchResult
from ax.utils.common.result import Err, Ok
from pandas import DataFrame
from pyre_extensions import none_throws


def _get_no_metadata_msg(trial_index: int) -> str:
    return f"No metadata available for trial {trial_index}."


class BenchmarkMetricBase(Metric):
    def __init__(
        self,
        name: str,
        # Needed to be boolean (not None) for validation of MOO opt configs
        lower_is_better: bool,
        observe_noise_sd: bool = True,
    ) -> None:
        """
        Args:
            name: Name of the metric.
            lower_is_better: If `True`, lower metric values are considered better.
            observe_noise_sd: If `True`, the standard deviation of the observation
                noise is included in the `sem` column of the the returned data.
                If `False`, `sem` is set to `None` (meaning that the model will
                have to infer the noise level).
        """
        super().__init__(name=name, lower_is_better=lower_is_better)
        # Declare `lower_is_better` as bool (rather than optional as in the base class)
        self.lower_is_better: bool = lower_is_better
        self.observe_noise_sd: bool = observe_noise_sd

    def _class_specific_metdata_validation(
        self, metadata: BenchmarkTrialMetadata | None
    ) -> None:
        return

    def fetch_trial_data(self, trial: BaseTrial, **kwargs: Any) -> MetricFetchResult:
        """
        Args:
            trial: The trial from which to fetch data.
            kwargs: Unsupported and will raise an exception.

        Returns:
            A MetricFetchResult containing the data for the requested metric.
        """

        class_name = self.__class__.__name__
        if len(kwargs) > 0:
            raise NotImplementedError(
                f"Arguments {set(kwargs)} are not supported in "
                f"{class_name}.fetch_trial_data."
            )
        if isinstance(trial, BatchTrial) and len(trial.abandoned_arms) > 0:
            raise NotImplementedError(
                f"{self.__class__.__name__} does not support abandoned arms in "
                "batch trials."
            )
        if len(trial.run_metadata) == 0:
            return Err(
                MetricFetchE(
                    message=_get_no_metadata_msg(trial_index=trial.index),
                    exception=None,
                )
            )

        metadata = trial.run_metadata["benchmark_metadata"]
        self._class_specific_metdata_validation(metadata=metadata)
        backend_simulator = metadata.backend_simulator
        df = metadata.dfs[self.name]

        # Filter out the observable data
        if backend_simulator is None:
            # If there's no backend simulator then no filtering is needed; the
            # trial will complete immediately, with all data available.
            available_data = df
        else:
            sim_trial = none_throws(
                backend_simulator.get_sim_trial_by_index(trial.index)
            )
            # The BackendSimulator distinguishes between queued and running
            # trials "for testing particular initialization cases", but these
            # are all "running" to orchestrator.
            start_time = none_throws(sim_trial.sim_start_time)

            if sim_trial.sim_completed_time is None:  # Still running
                max_t = backend_simulator.time - start_time
            elif sim_trial.sim_completed_time > backend_simulator.time:
                raise RuntimeError(
                    "The trial's completion time is in the future! This is "
                    f"unexpected. {sim_trial.sim_completed_time=}, "
                    f"{backend_simulator.time=}"
                )
            else:
                # Completed, may have stopped early -- can't assume all data available
                completed_time = none_throws(sim_trial.sim_completed_time)
                max_t = completed_time - start_time

            available_data = df[df["virtual runtime"] <= max_t]

        if not self.observe_noise_sd:
            available_data.loc[:, "sem"] = np.nan
        return self._df_to_result(df=available_data.drop(columns=["virtual runtime"]))

    @abstractmethod
    def _df_to_result(self, df: DataFrame) -> MetricFetchResult:
        """
        Convert a DataFrame of observable data to Data or MapData, as
        appropriate for the class.
        """
        ...


class BenchmarkMetric(BenchmarkMetricBase):
    """
    Non-map Metric for benchmarking that is not available while running.

    It cannot process data with multiple time steps, as it would only return one
    value -- the value it has at completion time -- regardless.
    """

    def _class_specific_metdata_validation(
        self, metadata: BenchmarkTrialMetadata | None
    ) -> None:
        if metadata is not None:
            df = metadata.dfs[self.name]
            if df["step"].nunique() > 1:
                raise ValueError(
                    f"Trial has data from multiple time steps. This is"
                    f" not supported by `{self.__class__.__name__}`; use "
                    "`BenchmarkMapMetric`."
                )

    def _df_to_result(self, df: DataFrame) -> MetricFetchResult:
        return Ok(value=Data(df=df.drop(columns=["step"])))


class BenchmarkTimeVaryingMetric(BenchmarkMetricBase):
    """
    Non-Map Metric for benchmarking that is available while running.

    It can produce different values at different times depending on when it is
    called, using the `time` on a `BackendSimulator`.
    """

    @classmethod
    def is_available_while_running(cls) -> bool:
        return True

    def _df_to_result(self, df: DataFrame) -> MetricFetchResult:
        return Ok(
            value=Data(df=df[df["step"] == df["step"].max()].drop(columns=["step"]))
        )


class BenchmarkMapMetric(MapMetric, BenchmarkMetricBase):
    """MapMetric for benchmarking. It is available while running."""

    @classmethod
    def is_available_while_running(cls) -> bool:
        return True

    def _df_to_result(self, df: DataFrame) -> MetricFetchResult:
        return Ok(value=MapData(df=df))


class BenchmarkMapUnavailableWhileRunningMetric(MapMetric, BenchmarkMetricBase):
    def _df_to_result(self, df: DataFrame) -> MetricFetchResult:
        return Ok(value=MapData(df=df))



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from __future__ import annotations

from collections.abc import Callable
from logging import Logger
from typing import TYPE_CHECKING

import numpy as np
import numpy.typing as npt
from ax.adapter.data_utils import ExperimentData
from ax.adapter.transforms.relativize import BaseRelativize, get_metric_index
from ax.core.observation import Observation, ObservationData, ObservationFeatures
from ax.core.optimization_config import OptimizationConfig
from ax.core.outcome_constraint import OutcomeConstraint
from ax.core.search_space import SearchSpace
from ax.core.utils import get_target_trial_index
from ax.generators.types import TConfig
from ax.utils.common.logger import get_logger
from ax.utils.stats.statstools import relativize, unrelativize
from pyre_extensions import assert_is_instance, none_throws

if TYPE_CHECKING:
    # import as module to make sphinx-autodoc-typehints happy
    from ax import adapter as adapter_module  # noqa F401
logger: Logger = get_logger(__name__)


class TransformToNewSQ(BaseRelativize):
    """Map relative values of one batch to SQ of another.

    Will compute the relative metrics for each arm in each batch, and will then turn
    those back into raw metrics but using the status quo values set on the Adapter.

    This is useful if batches are comparable on a relative scale, but
    have offset in their status quo. This is often approximately true for online
    experiments run in separate batches.

    Note that relativization is done using the delta method, so it will not
    simply be the ratio of the means.
    """

    def __init__(
        self,
        search_space: SearchSpace | None = None,
        observations: list[Observation] | None = None,
        experiment_data: ExperimentData | None = None,
        adapter: adapter_module.base.Adapter | None = None,
        config: TConfig | None = None,
    ) -> None:
        """Initialize the transform.

        Args:
            config: Can be used to specify the target trial index. The SQ data from
                the target trial will be used to relativize the data from all other
                trials. If not specified, the target trial will be inferred using
                `get_target_trial_index`.
        """
        super().__init__(
            search_space=search_space,
            observations=observations,
            experiment_data=experiment_data,
            adapter=adapter,
            config=config,
        )
        self._status_quo_name: str = none_throws(none_throws(adapter).status_quo_name)

        target_trial_index = None

        if config is not None:
            target_trial_index = config.get("target_trial_index", None)

        if (
            target_trial_index is None
            and adapter is not None
            and adapter._experiment is not None
        ):
            target_trial_index = get_target_trial_index(experiment=adapter._experiment)
            trials_indices_with_sq_data = self.status_quo_data_by_trial.keys()
            if target_trial_index not in trials_indices_with_sq_data:
                target_trial_index = max(trials_indices_with_sq_data)
                logger.warning(
                    "No status quo data for target trial. Failing back to "
                    f"{target_trial_index}."
                )

        if target_trial_index is not None:
            self.default_trial_idx: int = assert_is_instance(
                target_trial_index,
                int,
            )

    @property
    def control_as_constant(self) -> bool:
        """Whether or not the control is treated as a constant in the model."""
        return True

    def transform_optimization_config(
        self,
        optimization_config: OptimizationConfig,
        adapter: adapter_module.base.Adapter | None = None,
        fixed_features: ObservationFeatures | None = None,
    ) -> OptimizationConfig:
        return optimization_config

    def untransform_outcome_constraints(
        self,
        outcome_constraints: list[OutcomeConstraint],
        fixed_features: ObservationFeatures | None = None,
    ) -> list[OutcomeConstraint]:
        return outcome_constraints

    def transform_experiment_data(
        self, experiment_data: ExperimentData
    ) -> ExperimentData:
        observation_data = experiment_data.observation_data.copy(deep=True)
        all_trial_indices = observation_data.index.get_level_values(
            "trial_index"
        ).unique()
        if not all_trial_indices.isin(self.status_quo_data_by_trial.keys()).all():
            raise ValueError(
                f"{self.__class__.__name__} requires status quo data for all "
                f"trials in the experiment data. Found trial indices "
                f"{all_trial_indices} but status quo data is only available for "
                f"trials {list(self.status_quo_data_by_trial.keys())}."
            )

        trial_indices = observation_data.index.get_level_values("trial_index")
        transform_mask = trial_indices != self.default_trial_idx
        # Get the target trial's status quo data
        target_sq_data = self.status_quo_data_by_trial[self.default_trial_idx]

        metrics = experiment_data.metric_names
        if not transform_mask.any():
            # Nothing to transform, set metrics to empty list to skip the loop.
            # We still need to drop SQ after.
            metrics = []

        for metric in metrics:
            # Create arrays of control values for each row based on trial_index.
            mean_c, sem_c = [], []
            for idx in trial_indices:
                sq_data = self.status_quo_data_by_trial[idx]
                j = get_metric_index(data=sq_data, metric_name=metric)
                mean_c.append(sq_data.means[j])
                sem_c.append(sq_data.covariance[j, j] ** 0.5)
            mean_c = np.array(mean_c)
            sem_c = np.array(sem_c)

            # Only transform rows that are not from the target trial.
            means_t = observation_data.loc[transform_mask, ("mean", metric)]
            sems_t = observation_data.loc[transform_mask, ("sem", metric)]

            # Relativize with respect to original trial's status quo.
            means_rel, sems_rel = relativize(
                means_t=means_t,
                sems_t=sems_t,
                mean_c=mean_c[transform_mask],
                sem_c=sem_c[transform_mask],
                as_percent=False,
                control_as_constant=self.control_as_constant,
            )

            # Unrelativize with respect to target trial's status quo.
            target_j = get_metric_index(data=target_sq_data, metric_name=metric)
            target_mean_c = target_sq_data.means[target_j]
            abs_target_mean_c = np.abs(target_mean_c)
            observation_data.loc[transform_mask, ("mean", metric)] = (
                means_rel * abs_target_mean_c + target_mean_c
            )
            observation_data.loc[transform_mask, ("sem", metric)] = (
                sems_rel * abs_target_mean_c
            )

        # Drop SQ observations from the data -- except for the target trial.
        # Keep rows where arm_name != SQ or trial_index == target_trial_idx.
        observation_data = observation_data[
            (
                observation_data.index.get_level_values("arm_name")
                != self._status_quo_name
            )
            | (
                observation_data.index.get_level_values("trial_index")
                == self.default_trial_idx
            )
        ]
        arm_data = experiment_data.arm_data
        arm_data = arm_data[
            (arm_data.index.get_level_values("arm_name") != self._status_quo_name)
            | (arm_data.index.get_level_values("trial_index") == self.default_trial_idx)
        ]
        return ExperimentData(
            arm_data=arm_data,
            observation_data=observation_data,
        )

    def _get_relative_data_from_obs(
        self,
        obs: Observation,
        rel_op: Callable[..., tuple[npt.NDArray, npt.NDArray]],
    ) -> ObservationData:
        idx = (
            int(obs.features.trial_index)
            if obs.features.trial_index is not None
            else self.default_trial_idx
        )
        if idx == self.default_trial_idx:
            # don't transform data from target batch
            return obs.data
        return super()._get_relative_data_from_obs(
            obs=obs,
            rel_op=rel_op,
        )

    def _rel_op_on_observations(
        self,
        observations: list[Observation],
        rel_op: Callable[..., tuple[npt.NDArray, npt.NDArray]],
    ) -> list[Observation]:
        rel_observations = super()._rel_op_on_observations(
            observations=observations,
            rel_op=rel_op,
        )
        return [
            obs
            for obs in rel_observations
            # drop SQ observations
            if (
                obs.arm_name != self._status_quo_name
                or obs.features.trial_index == self.default_trial_idx
            )
        ]

    def _get_rel_mean_sem(
        self,
        means_t: float,
        sems_t: float,
        mean_c: float,
        sem_c: float,
        metric: str,
        rel_op: Callable[..., tuple[npt.NDArray, npt.NDArray]],
    ) -> tuple[float, float]:
        """Compute (un)transformed mean and sem for a single metric."""
        target_status_quo_data = self.status_quo_data_by_trial[self.default_trial_idx]
        j = get_metric_index(data=target_status_quo_data, metric_name=metric)
        target_mean_c = target_status_quo_data.means[j]
        abs_target_mean_c = np.abs(target_mean_c)
        if rel_op == unrelativize:
            means_t = (means_t - target_mean_c) / abs_target_mean_c
            sems_t = sems_t / abs_target_mean_c
        means_rel, sems_rel = rel_op(
            means_t=means_t,
            sems_t=sems_t,
            mean_c=mean_c,
            sem_c=sem_c,
            as_percent=False,
            control_as_constant=self.control_as_constant,
        )
        if rel_op == relativize:
            means_rel = means_rel * abs_target_mean_c + target_mean_c
            sems_rel = sems_rel * abs_target_mean_c
        # pyre-fixme[7]: Expected `Tuple[float, float]` but got
        #  `Tuple[ndarray[typing.Any, typing.Any], ndarray[typing.Any, typing.Any]]`.
        return means_rel, sems_rel



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

import itertools
import logging
from collections.abc import Sequence
from functools import reduce
from operator import mul

import numpy.typing as npt
from ax.core.types import TGenMetadata, TParamValue, TParamValueList
from ax.generators.discrete_base import DiscreteGenerator
from ax.generators.types import TConfig
from ax.utils.common.docutils import copy_doc
from ax.utils.common.logger import get_logger


logger: logging.Logger = get_logger(__name__)


class FullFactorialGenerator(DiscreteGenerator):
    """Generator for full factorial designs.

    Generates arms for all possible combinations of parameter values,
    each with weight 1.

    The value of n supplied to `gen` will be ignored, as the number
    of arms generated is determined by the list of parameter values.
    To suppress this warning, use n = -1.
    """

    def __init__(
        self, max_cardinality: int = 100, check_cardinality: bool = True
    ) -> None:
        """
        Args:
            max_cardinality: maximum number of arms allowed if
                check_cardinality == True. Default is 100.
            check_cardinality: if True, throw if number of arms
                exceeds max_cardinality.
        """
        super().__init__()
        self.max_cardinality = max_cardinality
        self.check_cardinality = check_cardinality

    @copy_doc(DiscreteGenerator.gen)
    # pyre-fixme[15]: Inconsistent override in return
    def gen(
        self,
        n: int,
        parameter_values: Sequence[Sequence[TParamValue]],
        objective_weights: npt.NDArray | None,
        outcome_constraints: tuple[npt.NDArray, npt.NDArray] | None = None,
        fixed_features: dict[int, TParamValue] | None = None,
        pending_observations: Sequence[Sequence[Sequence[TParamValue]]] | None = None,
        model_gen_options: TConfig | None = None,
    ) -> tuple[list[TParamValueList], list[float], TGenMetadata]:
        if fixed_features:
            # Make a copy so as to not mutate it
            parameter_values = list(parameter_values)
            for fixed_feature_index, fixed_feature_value in fixed_features.items():
                parameter_values[fixed_feature_index] = [fixed_feature_value]

        num_arms = reduce(mul, [len(values) for values in parameter_values], 1)
        if n != num_arms:
            logger.warning(
                "FullFactorialGenerator will ignore the specified value of n. "
                "The generator automatically determines how many arms to "
                "generate."
            )
        if self.check_cardinality and num_arms > self.max_cardinality:
            raise ValueError(
                f"FullFactorialGenerator generated {num_arms} arms, "
                f"but the maximum number of arms allowed is "
                f"{self.max_cardinality}."
            )

        points = [list(x) for x in itertools.product(*parameter_values)]
        return (points, [1.0 for _ in range(len(points))], {})



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from __future__ import annotations

import math
import operator
from collections.abc import Callable
from functools import partial, reduce
from itertools import product
from logging import Logger
from typing import Any, Mapping, Sequence

import torch
from ax.core.search_space import SearchSpaceDigest
from ax.exceptions.core import AxError, DataRequiredError, SearchSpaceExhausted
from ax.generators.model_utils import (
    all_ordinal_features_are_integer_valued,
    enumerate_discrete_combinations,
    mk_discrete_choices,
)
from ax.generators.torch.botorch_modular.optimizer_argparse import optimizer_argparse
from ax.generators.torch.botorch_modular.surrogate import Surrogate
from ax.generators.torch.botorch_moo_defaults import infer_objective_thresholds
from ax.generators.torch.utils import (
    _get_X_pending_and_observed,
    get_botorch_objective_and_transform,
    subset_model,
)
from ax.generators.torch_base import TorchOptConfig
from ax.utils.common.base import Base
from ax.utils.common.constants import Keys
from ax.utils.common.logger import get_logger
from botorch.acquisition.acquisition import AcquisitionFunction
from botorch.acquisition.input_constructors import get_acqf_input_constructor
from botorch.acquisition.knowledge_gradient import qKnowledgeGradient
from botorch.acquisition.multioutput_acquisition import MultiOutputAcquisitionFunction
from botorch.acquisition.objective import MCAcquisitionObjective, PosteriorTransform
from botorch.acquisition.risk_measures import RiskMeasureMCObjective
from botorch.exceptions.errors import BotorchError, InputDataError
from botorch.models.model import Model
from botorch.optim.optimize import (
    optimize_acqf,
    optimize_acqf_discrete,
    optimize_acqf_discrete_local_search,
    optimize_acqf_mixed,
)
from botorch.optim.optimize_mixed import optimize_acqf_mixed_alternating
from botorch.utils.constraints import get_outcome_constraint_transforms
from pyre_extensions import none_throws
from torch import Tensor

try:
    from botorch.utils.multi_objective.optimize import optimize_with_nsgaii
except ImportError:
    optimize_with_nsgaii = None


logger: Logger = get_logger(__name__)


# For fully discrete search spaces.
MAX_CHOICES_ENUMERATE = 100_000
MAX_CARDINALITY_FOR_LOCAL_SEARCH = 100
# For mixed search spaces.
ALTERNATING_OPTIMIZER_THRESHOLD = 10


def determine_optimizer(
    search_space_digest: SearchSpaceDigest,
    acqf: AcquisitionFunction | None = None,
    discrete_choices: Mapping[int, Sequence[float]] | None = None,
) -> str:
    """Determine the optimizer to use for a given search space.

    Args:
        search_space_digest: A SearchSpaceDigest object containing search space
            properties, e.g. ``bounds`` for optimization.
        acqf: The acquisition function to be used.
        discrete_choices: A dictionary mapping indices of discrete (ordinal
            or categorical) parameters to their respective sets of values
            provided as a list. This excludes fixed features.
    Returns:
        The name of the optimizer to use for the given search space.
    """
    if acqf is not None and isinstance(acqf, MultiOutputAcquisitionFunction):
        return "optimize_with_nsgaii"
    ssd = search_space_digest
    discrete_features = sorted(ssd.ordinal_features + ssd.categorical_features)
    if discrete_choices is None:
        discrete_choices = {}

    if len(discrete_features) == 0:
        optimizer = "optimize_acqf"
    else:
        fully_discrete = len(discrete_choices) == len(ssd.feature_names)
        if fully_discrete:
            # One of the three optimizers may be used depending on the number of
            # discrete choices and the cardinality of individual parameters.
            # If there are less than `MAX_CHOICES_ENUMERATE` choices, we will
            # evaluate all of them and pick the best.
            # If there are less than `MAX_CARDINALITY_FOR_LOCAL_SEARCH` choices
            # for all parameters or if any of the parameters has non-integer
            # valued choices, we will use local search. Otherwise, we will use
            # the mixed alternating optimizer, which may use continuous relaxation
            # for the high cardinality parameters, while using local search for
            # the remaining parameters.
            cardinalities = [len(c) for c in discrete_choices.values()]
            max_cardinality = max(cardinalities)
            total_discrete_choices = reduce(operator.mul, cardinalities)
            if total_discrete_choices > MAX_CHOICES_ENUMERATE:
                if (
                    max_cardinality <= MAX_CARDINALITY_FOR_LOCAL_SEARCH
                    or not all_ordinal_features_are_integer_valued(ssd=ssd)
                ):
                    optimizer = "optimize_acqf_discrete_local_search"
                else:
                    optimizer = "optimize_acqf_mixed_alternating"
            else:
                optimizer = "optimize_acqf_discrete"
        else:
            n_combos = math.prod([len(v) for v in discrete_choices.values()])
            # If there are
            # - any ordinal features with non-integer choices,
            # - less than `ALTERNATING_OPTIMIZER_THRESHOLD` combinations of
            #   discrete choices,
            # we will use `optimize_acqf_mixed`, which enumerates all discrete
            # combinations and optimizes the continuous features with discrete
            # features being fixed. Otherwise, we will use
            # `optimize_acqf_mixed_alternating`, which alternates between
            # continuous and discrete optimization steps.
            if (
                n_combos <= ALTERNATING_OPTIMIZER_THRESHOLD
                or not all_ordinal_features_are_integer_valued(ssd=ssd)
            ):
                optimizer = "optimize_acqf_mixed"
            else:
                optimizer = "optimize_acqf_mixed_alternating"
    return optimizer


class Acquisition(Base):
    """
    **All classes in 'botorch_modular' directory are under
    construction, incomplete, and should be treated as alpha
    versions only.**

    Ax wrapper for BoTorch `AcquisitionFunction`, subcomponent
    of `BoTorchGenerator` and is not meant to be used outside of it.

    Args:
        surrogate: The Surrogate model, with which this acquisition
            function will be used.
        search_space_digest: A SearchSpaceDigest object containing metadata
            about the search space (e.g. bounds, parameter types).
        torch_opt_config: A TorchOptConfig object containing optimization
            arguments (e.g., objective weights, constraints).
        botorch_acqf_class: Type of BoTorch `AcquisitionFunction` that
            should be used.
        botorch_acqf_options: Optional mapping of kwargs to the underlying
            `AcquisitionFunction` in BoTorch.
        botorch_acqf_classes_with_options: A list of tuples of botorch
            `AcquisitionFunction` classes and dicts of kwargs, passed to
            the botorch `AcquisitionFunction`. This is used to specify
            multiple acquisition functions to be used with MultiAcquisition.
        n: The number of candidates that will be generated by this acquisition.
        options: Optional mapping of kwargs to the underlying `Acquisition
            Function` in BoTorch.
    """

    surrogate: Surrogate
    acqf: AcquisitionFunction
    _model: Model
    _objective_weights: Tensor
    _objective_thresholds: Tensor | None
    _outcome_constraints: tuple[Tensor, Tensor] | None
    _risk_measure: RiskMeasureMCObjective | None
    _learned_objective_preference_model: Model | None
    _subset_idcs: Tensor | None

    def __init__(
        self,
        surrogate: Surrogate,
        search_space_digest: SearchSpaceDigest,
        torch_opt_config: TorchOptConfig,
        botorch_acqf_class: type[AcquisitionFunction] | None,
        botorch_acqf_options: dict[str, Any] | None = None,
        botorch_acqf_classes_with_options: list[
            tuple[type[AcquisitionFunction], dict[str, Any]]
        ]
        | None = None,
        n: int | None = None,
        options: dict[str, Any] | None = None,
    ) -> None:
        self.surrogate = surrogate
        options = options or {}
        botorch_acqf_options = botorch_acqf_options or {}
        self.search_space_digest = search_space_digest
        self.n = n
        self._should_subset_model: bool = options.get(Keys.SUBSET_MODEL, True)
        self._learned_objective_preference_model = None
        self._subset_idcs = None
        self._risk_measure = torch_opt_config.risk_measure

        # Extract pending and observed points.
        X_pending, X_observed = _get_X_pending_and_observed(
            Xs=surrogate.Xs,
            objective_weights=torch_opt_config.objective_weights,
            bounds=search_space_digest.bounds,
            pending_observations=torch_opt_config.pending_observations,
            outcome_constraints=torch_opt_config.outcome_constraints,
            linear_constraints=torch_opt_config.linear_constraints,
            fixed_features=torch_opt_config.fixed_features,
        )
        self.X_pending: Tensor | None = X_pending
        self.X_observed: Tensor | None = X_observed

        # Store objective thresholds for all outcomes (including non-objectives).
        self._full_objective_thresholds: Tensor | None = (
            torch_opt_config.objective_thresholds
        )
        self._full_objective_weights: Tensor = torch_opt_config.objective_weights

        (
            self._model,
            self._objective_weights,
            self._outcome_constraints,
            self._objective_thresholds,
        ) = self._subset_model(
            model=surrogate.model,
            objective_weights=torch_opt_config.objective_weights,
            outcome_constraints=torch_opt_config.outcome_constraints,
            objective_thresholds=torch_opt_config.objective_thresholds,
        )
        self._update_objective_thresholds(torch_opt_config=torch_opt_config)
        self._set_preference_model(torch_opt_config=torch_opt_config)

        if not botorch_acqf_classes_with_options:
            if botorch_acqf_class is None:
                raise AxError(
                    "One of botorch_acqf_class or botorch_acqf_classes_with_options"
                    " is required."
                )
            botorch_acqf_classes_with_options = [
                (botorch_acqf_class, botorch_acqf_options)
            ]
        self._instantiate_acquisition(
            botorch_acqf_classes_with_options=botorch_acqf_classes_with_options
        )

    def _subset_model(
        self,
        model: Model,
        objective_weights: Tensor,
        outcome_constraints: tuple[Tensor, Tensor] | None = None,
        objective_thresholds: Tensor | None = None,
    ) -> tuple[Model, Tensor, tuple[Tensor, Tensor] | None, Tensor | None]:
        if not self._should_subset_model:
            return model, objective_weights, outcome_constraints, objective_thresholds
        # Otherwise, subset
        subset_model_results = subset_model(
            model=model,
            objective_weights=objective_weights,
            outcome_constraints=outcome_constraints,
            objective_thresholds=objective_thresholds,
        )
        if self._subset_idcs is None:
            self._subset_idcs = subset_model_results.indices
        elif not torch.equal(subset_model_results.indices, self._subset_idcs):
            raise ValueError(
                "Subsequent model subsetting inconsistent with earlier."
            )  # pragma: no cover
        return (
            subset_model_results.model,
            subset_model_results.objective_weights,
            subset_model_results.outcome_constraints,
            subset_model_results.objective_thresholds,
        )

    def _update_objective_thresholds(self, torch_opt_config: TorchOptConfig) -> None:
        """If MOO and some objective thresholds are not specified, infer them using
        the model that has already been subset to avoid re-subsetting it within
        `infer_objective_thresholds`.

        If risk measures are used, objective thresholds must be provided. If not,
        this will error out.

        If `infer_objective_thresholds` errors out, e.g., due to no feasible point,
        this will log an error and let the optimization continue. Not all acquisition
        functions require objective thresholds, so this is not necessarily a problem.
        """
        if not (
            torch_opt_config.is_moo
            and (
                self._full_objective_thresholds is None
                or self._full_objective_thresholds[self._full_objective_weights != 0]
                .isnan()
                .any()
            )
            and self.X_observed is not None
        ):
            return
        if torch_opt_config.risk_measure is not None:
            raise NotImplementedError(
                "Objective thresholds must be provided when using risk measures."
            )
        try:
            self._full_objective_thresholds = infer_objective_thresholds(
                model=self._model,
                objective_weights=self._full_objective_weights,
                outcome_constraints=torch_opt_config.outcome_constraints,
                X_observed=self.X_observed,
                subset_idcs=self._subset_idcs,
                objective_thresholds=self._full_objective_thresholds,
            )
            self._objective_thresholds = (
                none_throws(self._full_objective_thresholds)[self._subset_idcs]
                if self._subset_idcs is not None
                else self._full_objective_thresholds
            )
        except (AxError, BotorchError) as e:
            logger.warning(
                "Failed to infer objective thresholds. Resuming optimization "
                "without objective thresholds, which may or may not work depending "
                f"on the acquisition function. Original error: {e}."
            )

    def _set_preference_model(self, torch_opt_config: TorchOptConfig) -> None:
        if torch_opt_config.use_learned_objective:
            if (Keys.PAIRWISE_PREFERENCE_QUERY.value,) not in self.surrogate._submodels:
                raise DataRequiredError(
                    "PreferenceOptimizationConfig is used but missing "
                    "preference objective model. Double check if the preference "
                    "exploration auxiliary experiment has data."
                )
            self._learned_objective_preference_model = self.surrogate._submodels[
                (Keys.PAIRWISE_PREFERENCE_QUERY.value,)
            ]

    def _instantiate_acquisition(
        self,
        botorch_acqf_classes_with_options: list[
            tuple[type[AcquisitionFunction], dict[str, Any]]
        ],
    ) -> None:
        """Constructs the acquisition function based on the provided
        botorch_acqf_classes_with_options.

        Args:
            botorch_acqf_classes: A list of BoTorch acquisition function classes.
        """
        if len(botorch_acqf_classes_with_options) != 1:
            raise ValueError("Only one botorch_acqf_class is supported.")
        botorch_acqf_class, botorch_acqf_options = botorch_acqf_classes_with_options[0]
        self.acqf = self._construct_botorch_acquisition(
            botorch_acqf_class=botorch_acqf_class,
            botorch_acqf_options=botorch_acqf_options,
            model=self._model,
        )
        self.models_used = [self.surrogate.model_name_by_metric]
        self.acq_function_sequence = None

    def _construct_botorch_acquisition(
        self,
        botorch_acqf_class: type[AcquisitionFunction],
        botorch_acqf_options: dict[str, Any],
        model: Model,
    ) -> AcquisitionFunction:
        objective, posterior_transform = self.get_botorch_objective_and_transform(
            botorch_acqf_class=botorch_acqf_class,
            model=model,
            objective_weights=self._objective_weights,
            objective_thresholds=self._objective_thresholds,
            outcome_constraints=self._outcome_constraints,
            X_observed=self.X_observed,
            risk_measure=self._risk_measure,
            learned_objective_preference_model=self._learned_objective_preference_model,
        )
        input_constructor_kwargs = {
            "model": model,
            "X_baseline": self.X_observed,
            "X_pending": self.X_pending,
            "objective_thresholds": self._objective_thresholds,
            "constraints": get_outcome_constraint_transforms(
                outcome_constraints=self._outcome_constraints
            ),
            "constraints_tuple": self._outcome_constraints,
            "objective": objective,
            "posterior_transform": posterior_transform,
            **botorch_acqf_options,
        }
        target_fidelities = {
            k: v
            for k, v in self.search_space_digest.target_values.items()
            if k in self.search_space_digest.fidelity_features
        }
        if len(target_fidelities) > 0:
            input_constructor_kwargs["target_fidelities"] = target_fidelities
        input_constructor = get_acqf_input_constructor(botorch_acqf_class)

        # Extract the training data from the surrogate.
        # If there is a single dataset, this will be the dataset itself.
        # If there are multiple datasets, this will be a dict mapping the outcome names
        # to the corresponding datasets.
        training_data = self.surrogate.training_data
        if len(training_data) == 1:
            training_data = training_data[0]
        else:
            training_data = dict(
                zip(none_throws(self.surrogate._outcomes), training_data)
            )
        acqf_inputs = input_constructor(
            training_data=training_data,
            bounds=self.search_space_digest.bounds,
            **{k: v for k, v in input_constructor_kwargs.items() if v is not None},
        )
        return botorch_acqf_class(**acqf_inputs)  # pyre-ignore [45]

    @property
    def botorch_acqf_class(self) -> type[AcquisitionFunction]:
        """BoTorch ``AcquisitionFunction`` class underlying this ``Acquisition``."""
        return self.acqf.__class__

    @property
    def dtype(self) -> torch.dtype | None:
        """Torch data type of the tensors in the training data used in the model,
        of which this ``Acquisition`` is a subcomponent.
        """
        return self.surrogate.dtype

    @property
    def device(self) -> torch.device | None:
        """Torch device type of the tensors in the training data used in the model,
        of which this ``Acquisition`` is a subcomponent.
        """
        return self.surrogate.device

    @property
    def objective_thresholds(self) -> Tensor | None:
        """The objective thresholds for all outcomes.

        For non-objective outcomes, the objective thresholds are nans.
        """
        return self._full_objective_thresholds

    @property
    def objective_weights(self) -> Tensor | None:
        """The objective weights for all outcomes."""
        return self._full_objective_weights

    def optimize(
        self,
        n: int,
        search_space_digest: SearchSpaceDigest,
        inequality_constraints: list[tuple[Tensor, Tensor, float]] | None = None,
        fixed_features: dict[int, float] | None = None,
        rounding_func: Callable[[Tensor], Tensor] | None = None,
        optimizer_options: dict[str, Any] | None = None,
    ) -> tuple[Tensor, Tensor, Tensor]:
        """Generate a set of candidates via multi-start optimization. Obtains
        candidates and their associated acquisition function values.

        Args:
            n: The number of candidates to generate.
            search_space_digest: A ``SearchSpaceDigest`` object containing search space
                properties, e.g. ``bounds`` for optimization.
            inequality_constraints: A list of tuples (indices, coefficients, rhs),
                with each tuple encoding an inequality constraint of the form
                ``sum_i (X[indices[i]] * coefficients[i]) >= rhs``.
            fixed_features: A map `{feature_index: value}` for features that
                should be fixed to a particular value during generation.
            rounding_func: A function that post-processes an optimization
                result appropriately. This is typically passed down from
                `Adapter` to ensure compatibility of the candidates with
                with Ax transforms. For additional post processing, use
                `post_processing_func` option in `optimizer_options`.
            optimizer_options: Options for the optimizer function, e.g. ``sequential``
                or ``raw_samples``. This can also include a `post_processing_func`
                which is applied to the candidates before the `rounding_func`.
                `post_processing_func` can be used to support more customized options
                that typically only exist in MBM, such as BoTorch transforms.
                See the docstring of `TorchOptConfig` for more information on passing
                down these options while constructing a generation strategy.

        Returns:
            A three-element tuple containing an `n x d`-dim tensor of generated
            candidates, a tensor with the associated acquisition values, and a tensor
            with the weight for each candidate.
        """
        # Options that would need to be passed in the transformed space are
        # disallowed, since this would be very difficult for an end user to do
        # directly, and someone who uses BoTorch at this level of detail would
        # probably be better off using BoTorch directly.
        # `return_best_only` and `return_full_tree` are disallowed because
        # Ax expects `optimize_acqf` to return tensors of a certain shape.
        if optimizer_options is not None:
            forbidden_optimizer_options = [
                "equality_constraints",
                "inequality_constraints",  # These should be constructed by Ax
                "batch_initial_conditions",
                "return_best_only",
                "return_full_tree",
            ]

            for kw in optimizer_options:
                if kw in forbidden_optimizer_options:
                    raise ValueError(
                        f"Argument {kw} is not allowed in `optimizer_options`."
                    )

        _tensorize = partial(torch.tensor, dtype=self.dtype, device=self.device)
        ssd = search_space_digest
        bounds = _tensorize(ssd.bounds).t()
        discrete_choices = mk_discrete_choices(ssd=ssd, fixed_features=fixed_features)

        optimizer = determine_optimizer(
            search_space_digest=ssd,
            discrete_choices=discrete_choices,
            acqf=self.acqf,
        )
        # `raw_samples` and `num_restarts` are not supported by
        # `optimize_acqf_discrete`.
        if (
            optimizer in ("optimize_acqf_discrete", "optimize_with_nsgaii")
            and optimizer_options is not None
        ):
            optimizer_options.pop("raw_samples", None)
            optimizer_options.pop("num_restarts", None)

        # Prepare arguments for optimizer
        optimizer_options_with_defaults = optimizer_argparse(
            self.acqf,
            optimizer_options=optimizer_options,
            optimizer=optimizer,
        )
        if fixed_features is not None:
            for i in fixed_features:
                if not 0 <= i < len(ssd.feature_names):
                    raise ValueError(f"Invalid fixed_feature index: {i}")
        # Return a weight of 1 for each arm by default. This can be
        # customized in subclasses if necessary.
        arm_weights = torch.ones(n, dtype=self.dtype)
        # 1. Handle the fully continuous search space.
        if optimizer == "optimize_acqf":
            candidates, acqf_values = optimize_acqf(
                acq_function=self.acqf,
                bounds=bounds,
                q=n,
                inequality_constraints=inequality_constraints,
                fixed_features=fixed_features,
                post_processing_func=rounding_func,
                acq_function_sequence=self.acq_function_sequence,
                **optimizer_options_with_defaults,
            )
            return candidates, acqf_values, arm_weights

        # 2. Handle fully discrete search spaces.
        if optimizer in (
            "optimize_acqf_discrete",
            "optimize_acqf_discrete_local_search",
        ):
            X_observed = self.X_observed
            if self.X_pending is not None:
                if X_observed is None:
                    X_observed = self.X_pending
                else:
                    X_observed = torch.cat([X_observed, self.X_pending], dim=0)

            # Special handling for search spaces with a large number of choices
            if optimizer == "optimize_acqf_discrete_local_search":
                discrete_choices = [
                    torch.tensor(c, device=self.device, dtype=self.dtype)
                    for c in discrete_choices.values()
                ]
                candidates, acqf_values = optimize_acqf_discrete_local_search(
                    acq_function=self.acqf,
                    q=n,
                    discrete_choices=discrete_choices,
                    inequality_constraints=inequality_constraints,
                    X_avoid=X_observed,
                    **optimizer_options_with_defaults,
                )
                return candidates, acqf_values, arm_weights

            # Else, optimizer is `optimize_acqf_discrete`
            # Enumerate all possible choices
            all_choices = (discrete_choices[i] for i in range(len(discrete_choices)))
            all_choices = _tensorize(tuple(product(*all_choices)))
            try:
                max_batch_size = optimizer_options_with_defaults.pop(
                    "max_batch_size", 2048
                )
                try:
                    # Adapt max batch size for batched models to reduce peak memory.
                    max_batch_size = (
                        max_batch_size // self.surrogate.model.batch_shape.numel()
                    )
                except Exception:  # pragma: no cover
                    pass  # Do not error out if the model does not have a batch shape.
                candidates, acqf_values = optimize_acqf_discrete(
                    acq_function=self.acqf,
                    q=n,
                    choices=all_choices,
                    max_batch_size=max_batch_size,
                    X_avoid=X_observed,
                    inequality_constraints=inequality_constraints,
                    **optimizer_options_with_defaults,
                )
            except InputDataError:
                raise SearchSpaceExhausted(
                    "No more feasible choices in a fully discrete search space."
                )
            return candidates, acqf_values, arm_weights

        # 3. Handle mixed search spaces that have discrete and continuous features.
        if optimizer == "optimize_acqf_mixed":
            candidates, acqf_values = optimize_acqf_mixed(
                acq_function=self.acqf,
                bounds=bounds,
                q=n,
                fixed_features_list=enumerate_discrete_combinations(
                    discrete_choices=discrete_choices
                ),
                inequality_constraints=inequality_constraints,
                post_processing_func=rounding_func,
                **optimizer_options_with_defaults,
            )
        elif optimizer == "optimize_acqf_mixed_alternating":
            candidates, acqf_values = optimize_acqf_mixed_alternating(
                acq_function=self.acqf,
                bounds=bounds,
                discrete_dims={
                    k: list(v)
                    for k, v in discrete_choices.items()
                    if k in search_space_digest.ordinal_features
                },
                cat_dims={
                    k: list(v)
                    for k, v in discrete_choices.items()
                    if k in search_space_digest.categorical_features
                },
                q=n,
                post_processing_func=rounding_func,
                fixed_features=fixed_features,
                inequality_constraints=inequality_constraints,
                **optimizer_options_with_defaults,
            )
        elif optimizer == "optimize_with_nsgaii":
            if optimize_with_nsgaii is not None:
                # TODO: support post_processing_func
                candidates, acqf_values = optimize_with_nsgaii(
                    acq_function=self.acqf,
                    bounds=bounds,
                    q=n,
                    fixed_features=fixed_features,
                    # We use pyre-ignore here to avoid a circular import.
                    # pyre-ignore [6]: Incompatible parameter type [6]: In call `len`,
                    # for 1st positional argument, expected
                    # `pyre_extensions.PyreReadOnly[Sized]` but got `Union[Tensor,
                    # Module]`.
                    num_objectives=len(self.acqf.acqfs),
                    **optimizer_options_with_defaults,
                )
                # It is possible that NSGA-II will return less than the requested
                # number of arms
                arm_weights = torch.ones(candidates.shape[0], dtype=self.dtype)
            else:
                raise AxError(
                    "optimize_with_nsgaii requires botorch to be installed with "
                    "the pymoo."
                )
        else:
            raise AxError(  # pragma: no cover
                f"Unknown optimizer: {optimizer}. This code should be unreachable."
            )
        return candidates, acqf_values, arm_weights

    def evaluate(self, X: Tensor) -> Tensor:
        """Evaluate the acquisition function on the candidate set `X`.

        Args:
            X: A `batch_shape x q x d`-dim Tensor of t-batches with `q` `d`-dim design
                points each.

        Returns:
            A `batch_shape'`-dim Tensor of acquisition values at the given
            design points `X`, where `batch_shape'` is the broadcasted batch shape of
            model and input `X`.
        """
        if isinstance(self.acqf, qKnowledgeGradient):
            return self.acqf.evaluate(X=X)
        else:
            # NOTE: `AcquisitionFunction.__call__` calls `forward`,
            # so below is equivalent to `self.acqf.forward(X=X)`.
            return self.acqf(X=X)

    def get_botorch_objective_and_transform(
        self,
        botorch_acqf_class: type[AcquisitionFunction],
        model: Model,
        objective_weights: Tensor,
        objective_thresholds: Tensor | None = None,
        outcome_constraints: tuple[Tensor, Tensor] | None = None,
        X_observed: Tensor | None = None,
        risk_measure: RiskMeasureMCObjective | None = None,
        learned_objective_preference_model: Model | None = None,
    ) -> tuple[MCAcquisitionObjective | None, PosteriorTransform | None]:
        return get_botorch_objective_and_transform(
            botorch_acqf_class=botorch_acqf_class,
            model=model,
            objective_weights=objective_weights,
            outcome_constraints=outcome_constraints,
            X_observed=X_observed,
            risk_measure=risk_measure,
            learned_objective_preference_model=learned_objective_preference_model,
        )



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict


import numpy as np
import pandas as pd
from ax.core.map_data import MAP_KEY, MapData
from ax.core.tests.test_data import TestDataBase
from ax.exceptions.core import UnsupportedError
from ax.utils.common.testutils import TestCase


class TestMapData(TestDataBase):
    cls: type[MapData] = MapData


class MapDataTest(TestCase):
    def setUp(self) -> None:
        super().setUp()
        self.df = pd.DataFrame(
            [
                {
                    "arm_name": "0_0",
                    MAP_KEY: 0,
                    "mean": 3.0,
                    "sem": 0.3,
                    "trial_index": 0,
                    "metric_name": "a",
                },
                # repeated arm 0_0
                {
                    "arm_name": "0_0",
                    MAP_KEY: 0,
                    "mean": 2.0,
                    "sem": 0.2,
                    "trial_index": 1,
                    "metric_name": "a",
                },
                {
                    "arm_name": "0_0",
                    MAP_KEY: 0,
                    "mean": 1.8,
                    "sem": 0.3,
                    "trial_index": 1,
                    "metric_name": "b",
                },
                {
                    "arm_name": "0_1",
                    MAP_KEY: 0,
                    "mean": 4.0,
                    "sem": 0.6,
                    "trial_index": 1,
                    "metric_name": "a",
                },
                {
                    "arm_name": "0_1",
                    MAP_KEY: 0,
                    "mean": 3.7,
                    "sem": 0.5,
                    "trial_index": 1,
                    "metric_name": "b",
                },
                {
                    "arm_name": "0_1",
                    MAP_KEY: 1,
                    "mean": 0.5,
                    "sem": None,
                    "trial_index": 1,
                    "metric_name": "a",
                },
                {
                    "arm_name": "0_1",
                    MAP_KEY: 1,
                    "mean": 3.0,
                    "sem": None,
                    "trial_index": 1,
                    "metric_name": "b",
                },
            ]
        )

        self.mmd = MapData(df=self.df)

    def test_df(self) -> None:
        df = self.mmd.df
        self.assertEqual(set(df["trial_index"].drop_duplicates()), {0, 1})

    def test_init(self) -> None:
        # Initialize empty
        empty = MapData()
        self.assertTrue(empty.map_df.empty)

        # Check that the required columns include the map keys.
        self.assertEqual(
            empty.REQUIRED_COLUMNS.union([MAP_KEY]), empty.required_columns()
        )
        self.assertEqual(set(empty.map_df.columns), empty.required_columns())

    def test_from_evaluations(self) -> None:
        with self.assertRaisesRegex(
            UnsupportedError, "MapData.from_evaluations is not supported"
        ):
            MapData.from_evaluations(evaluations={}, trial_index=0)

    def test_combine(self) -> None:
        with self.subTest("From no MapDatas"):
            data = MapData.from_multiple_map_data([])
            self.assertEqual(data.map_df.size, 0)

        with self.subTest("From two MapDatas"):
            mmd_double = MapData.from_multiple_map_data([self.mmd, self.mmd])
            self.assertEqual(mmd_double.map_df.size, 2 * self.mmd.map_df.size)

    def test_from_map_evaluations(self) -> None:
        for sem in (0.5, None):
            eval1 = (3.7, sem) if sem is not None else 3.7
            eval2 = (3.8, sem) if sem is not None else 3.8
            map_data = MapData.from_map_evaluations(
                evaluations={"0_1": [(1.0, {"b": eval1}), (1.0, {"b": eval2})]},
                trial_index=0,
            )
            self.assertEqual(map_data.map_df["sem"].isnull().all(), sem is None)
            self.assertEqual(len(map_data.map_df), 2)

    def test_upcast(self) -> None:
        fresh = MapData(df=self.df)
        # Assert df is not cached before first call
        self.assertIsNone(fresh._memo_df)

        self.assertEqual(
            fresh.df.columns.size,
            fresh.map_df.columns.size,
        )

        # Assert df is cached after first call
        self.assertIsNotNone(fresh._memo_df)

        self.assertTrue(
            fresh.df.equals(
                fresh.map_df.sort_values(MAP_KEY).drop_duplicates(
                    MapData.DEDUPLICATE_BY_COLUMNS, keep="last"
                )
            )
        )

    def test_latest(self) -> None:
        seed = 8888

        arm_names = ["0_0", "1_0", "2_0", "3_0"]
        max_epochs = [25, 50, 75, 100]
        metric_names = ["a", "b"]
        large_map_df = pd.DataFrame(
            [
                {
                    "arm_name": arm_name,
                    MAP_KEY: epoch + 1,
                    "mean": epoch * 0.1,
                    "sem": 0.1,
                    "trial_index": trial_index,
                    "metric_name": metric_name,
                }
                for metric_name in metric_names
                for trial_index, (arm_name, max_epoch) in enumerate(
                    zip(arm_names, max_epochs)
                )
                for epoch in range(max_epoch)
            ]
        )
        large_map_data = MapData(df=large_map_df)

        shuffled_large_map_df = large_map_data.map_df.groupby(
            MapData.DEDUPLICATE_BY_COLUMNS
        ).sample(frac=1, random_state=seed)
        shuffled_large_map_data = MapData(df=shuffled_large_map_df)

        for rows_per_group in [1, 40]:
            large_map_data_latest = large_map_data.latest(rows_per_group=rows_per_group)

            if rows_per_group == 1:
                self.assertTrue(
                    large_map_data_latest.map_df.groupby("metric_name")[MAP_KEY]
                    .transform(lambda col: set(col) == set(max_epochs))
                    .all()
                )

            # when rows_per_group is larger than the number of rows
            # actually observed in a group
            actual_rows_per_group = large_map_data_latest.map_df.groupby(
                MapData.DEDUPLICATE_BY_COLUMNS
            ).size()
            expected_rows_per_group = np.minimum(
                large_map_data_latest.map_df.groupby(MapData.DEDUPLICATE_BY_COLUMNS)[
                    MAP_KEY
                ]
                .max()
                .astype(int),
                rows_per_group,
            )
            self.assertTrue(actual_rows_per_group.equals(expected_rows_per_group))

            # behavior should be consistent despite shuffling
            shuffled_large_map_data_latest = shuffled_large_map_data.latest(
                rows_per_group=rows_per_group
            )
            self.assertTrue(
                shuffled_large_map_data_latest.map_df.equals(
                    large_map_data_latest.map_df
                )
            )

    def test_subsample(self) -> None:
        arm_names = ["0_0", "1_0", "2_0", "3_0"]
        max_epochs = [25, 50, 75, 100]
        metric_names = ["a", "b"]
        large_map_df = pd.DataFrame(
            [
                {
                    "arm_name": arm_name,
                    MAP_KEY: epoch + 1,
                    "mean": epoch * 0.1,
                    "sem": 0.1,
                    "trial_index": trial_index,
                    "metric_name": metric_name,
                }
                for metric_name in metric_names
                for trial_index, (arm_name, max_epoch) in enumerate(
                    zip(arm_names, max_epochs)
                )
                for epoch in range(max_epoch)
            ]
        )
        large_map_data = MapData(df=large_map_df)
        large_map_df_sparse_metric = pd.DataFrame(
            [
                {
                    "arm_name": arm_name,
                    MAP_KEY: epoch + 1,
                    "mean": epoch * 0.1,
                    "sem": 0.1,
                    "trial_index": trial_index,
                    "metric_name": metric_name,
                }
                for metric_name in metric_names
                for trial_index, (arm_name, max_epoch) in enumerate(
                    zip(arm_names, max_epochs)
                )
                for epoch in range(max_epoch if metric_name == "a" else max_epoch // 5)
            ]
        )
        large_map_data_sparse_metric = MapData(df=large_map_df_sparse_metric)

        # test keep_every
        subsample = large_map_data.subsample(keep_every=10)
        self.assertEqual(len(subsample.map_df), 52)
        subsample = large_map_data.subsample(keep_every=25)
        self.assertEqual(len(subsample.map_df), 20)
        subsample = large_map_data.subsample(limit_rows_per_group=7)
        self.assertEqual(len(subsample.map_df), 36)

        # test limit_rows_per_group
        subsample = large_map_data.subsample(limit_rows_per_group=1)
        self.assertEqual(len(subsample.map_df), 8)
        subsample = large_map_data.subsample(limit_rows_per_group=7)
        self.assertEqual(len(subsample.map_df), 36)
        subsample = large_map_data.subsample(limit_rows_per_group=10)
        self.assertEqual(len(subsample.map_df), 52)
        subsample = large_map_data.subsample(limit_rows_per_group=1000)
        self.assertEqual(len(subsample.map_df), 500)

        # test limit_rows_per_metric
        subsample = large_map_data.subsample(limit_rows_per_metric=50)
        self.assertEqual(len(subsample.map_df), 100)
        subsample = large_map_data.subsample(limit_rows_per_metric=65)
        self.assertEqual(len(subsample.map_df), 128)
        subsample = large_map_data.subsample(limit_rows_per_metric=1000)
        self.assertEqual(len(subsample.map_df), 500)

        # test include_first_last
        subsample = large_map_data.subsample(
            limit_rows_per_metric=20, include_first_last=True
        )
        self.assertEqual(len(subsample.map_df), 40)
        # check that we 1 and 100 are included
        self.assertEqual(subsample.map_df[MAP_KEY].min(), 1)
        self.assertEqual(subsample.map_df[MAP_KEY].max(), 100)
        subsample = large_map_data.subsample(
            limit_rows_per_metric=20, include_first_last=False
        )
        self.assertEqual(len(subsample.map_df), 40)
        self.assertEqual(subsample.map_df[MAP_KEY].min(), 1)
        self.assertEqual(subsample.map_df[MAP_KEY].max(), 92)

        # test limit_rows_per_metric when some metrics are sparsely
        # reported (we shouldn't subsample those)
        subsample = large_map_data_sparse_metric.subsample(
            limit_rows_per_metric=100, include_first_last=False
        )
        map_df = large_map_data_sparse_metric.map_df
        subsample_map_df = subsample.map_df
        self.assertEqual(
            len(subsample_map_df[subsample_map_df["metric_name"] == "a"]), 85
        )
        self.assertEqual(
            len(subsample_map_df[subsample_map_df["metric_name"] == "b"]),
            len(map_df[map_df["metric_name"] == "b"]),
        )



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from ax.analysis.plotly.surface.contour import compute_contour_adhoc, ContourPlot
from ax.core.trial import Trial
from ax.exceptions.core import UserInputError
from ax.service.ax_client import AxClient, ObjectiveProperties
from ax.utils.common.testutils import TestCase
from ax.utils.testing.mock import mock_botorch_optimize

from pyre_extensions import assert_is_instance, none_throws


class TestContourPlot(TestCase):
    @mock_botorch_optimize
    def setUp(self) -> None:
        super().setUp()
        self.client = AxClient()
        self.client.create_experiment(
            is_test=True,
            name="foo",
            parameters=[
                {
                    "name": "x",
                    "type": "range",
                    "bounds": [-1.0, 1.0],
                },
                {
                    "name": "y",
                    "type": "range",
                    "bounds": [-1.0, 1.0],
                },
            ],
            objectives={"bar": ObjectiveProperties(minimize=True)},
        )

        for _ in range(10):
            parameterization, trial_index = self.client.get_next_trial()
            self.client.complete_trial(
                trial_index=trial_index,
                raw_data={
                    "bar": parameterization["x"] ** 2 + parameterization["y"] ** 2
                },
            )
        self.expected_subtitle = (
            "The contour plot visualizes the predicted outcomes "
            "for bar across a two-dimensional parameter space, "
            "with other parameters held fixed at their status_quo value "
            "(or mean value if status_quo is unavailable). This plot helps "
            "in identifying regions of optimal performance and understanding "
            "how changes in the selected parameters influence the predicted "
            "outcomes. Contour lines represent levels of constant predicted "
            "values, providing insights into the gradient and potential optima "
            "within the parameter space."
        )
        self.expected_title = "x, y vs. bar"
        self.expected_name = "ContourPlot"
        self.expected_cols = {
            "x",
            "y",
            "bar_mean",
            "bar_sem",
            "sampled",
            "trial_index",
            "arm_name",
        }

    def test_compute(self) -> None:
        analysis = ContourPlot(
            x_parameter_name="x", y_parameter_name="y", metric_name="bar"
        )

        # Test that it fails if no Experiment is provided
        with self.assertRaisesRegex(UserInputError, "requires an Experiment"):
            analysis.compute()
        # Test that it fails if no GenerationStrategy is provided
        with self.assertRaisesRegex(
            UserInputError, "Must provide either a GenerationStrategy or an Adapter"
        ):
            analysis.compute(experiment=self.client.experiment)

        card = analysis.compute(
            experiment=self.client.experiment,
            generation_strategy=self.client.generation_strategy,
        )
        self.assertEqual(
            card.name,
            self.expected_name,
        )
        self.assertEqual(card.title, self.expected_title)
        self.assertEqual(card.subtitle, self.expected_subtitle)
        self.assertEqual(
            {*card.df.columns},
            self.expected_cols,
        )
        self.assertIsNotNone(card.blob)

        # Assert that any row where sampled is True has a value of x that is
        # sampled in at least one trial.
        x_values_sampled = {
            none_throws(assert_is_instance(trial, Trial).arm).parameters["x"]
            for trial in self.client.experiment.trials.values()
        }
        y_values_sampled = {
            none_throws(assert_is_instance(trial, Trial).arm).parameters["y"]
            for trial in self.client.experiment.trials.values()
        }
        self.assertTrue(
            card.df.apply(
                lambda row: row["x"] in x_values_sampled
                and row["y"] in y_values_sampled
                if row["sampled"]
                else True,
                axis=1,
            ).all()
        )

        # Less-than-or-equal to because we may have removed some duplicates
        self.assertTrue(card.df["sampled"].sum() <= len(self.client.experiment.trials))

    def test_compute_adhoc(self) -> None:
        card = compute_contour_adhoc(
            x_parameter_name="x",
            y_parameter_name="y",
            metric_name="bar",
            experiment=self.client.experiment,
            generation_strategy=self.client.generation_strategy,
        )
        self.assertEqual(
            card.name,
            self.expected_name,
        )
        self.assertEqual(card.title, self.expected_title)
        self.assertEqual(card.subtitle, self.expected_subtitle)
        self.assertEqual({*card.df.columns}, self.expected_cols)
        self.assertIsNotNone(card.blob)

        # Assert that any row where sampled is True has a value of x that is
        # sampled in at least one trial.
        x_values_sampled = {
            none_throws(assert_is_instance(trial, Trial).arm).parameters["x"]
            for trial in self.client.experiment.trials.values()
        }
        y_values_sampled = {
            none_throws(assert_is_instance(trial, Trial).arm).parameters["y"]
            for trial in self.client.experiment.trials.values()
        }
        self.assertTrue(
            card.df.apply(
                lambda row: row["x"] in x_values_sampled
                and row["y"] in y_values_sampled
                if row["sampled"]
                else True,
                axis=1,
            ).all()
        )

        # Less-than-or-equal to because we may have removed some duplicates
        self.assertTrue(card.df["sampled"].sum() <= len(self.client.experiment.trials))

    def test_trial_status_filtering(self) -> None:
        trial_index = self.client.experiment.new_trial().index
        self.client.experiment.trials[trial_index].mark_abandoned()

        analysis = ContourPlot(
            x_parameter_name="x", y_parameter_name="y", metric_name="bar"
        )
        card = analysis.compute(
            experiment=self.client.experiment,
            generation_strategy=self.client.generation_strategy,
        )
        self.assertNotIn(
            trial_index,
            card.df["trial_index"].values,
        )



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from unittest.mock import Mock, patch

from ax.core.trial_status import TrialStatus
from ax.utils.common.logger import get_logger
from ax.utils.common.testutils import TestCase
from ax.utils.testing.backend_simulator import BackendSimulator, BackendSimulatorOptions
from ax.utils.testing.utils_testing_stubs import get_backend_simulator_with_trials


class BackendSimulatorTest(TestCase):
    @patch("ax.utils.testing.backend_simulator.time.time")
    def test_backend_simulator(self, time_mock: Mock) -> None:
        time_mock.return_value = 0.0
        dt = 0.001
        options = BackendSimulatorOptions(max_concurrency=2)

        # test init
        sim = BackendSimulator(options=options)
        options = sim.options
        self.assertEqual(options.max_concurrency, 2)
        self.assertEqual(options.time_scaling, 1.0)
        self.assertEqual(options.failure_rate, 0.0)
        self.assertEqual(sim.num_queued, 0)
        self.assertEqual(sim.num_running, 0)
        self.assertEqual(sim.num_failed, 0)
        self.assertEqual(sim.num_completed, 0)

        # test run trial
        sim.run_trial(0, dt)
        self.assertEqual(sim.num_queued, 0)
        self.assertEqual(sim.num_running, 1)
        sim.run_trial(1, dt)
        self.assertEqual(sim.num_queued, 0)
        self.assertEqual(sim.num_running, 2)
        sim.run_trial(2, dt)
        self.assertEqual(sim.num_queued, 1)
        self.assertEqual(sim.num_running, 2)
        status = sim.status()
        self.assertEqual(status.queued, [2])
        self.assertEqual(status.running, [0, 1])
        self.assertEqual(status.failed, [])
        self.assertEqual(status.completed, [])

        # "Wait" some time
        time_mock.return_value += 1.5 * dt
        sim.update()
        self.assertEqual(sim.num_queued, 0)
        self.assertEqual(sim.num_running, 1)
        self.assertEqual(sim.num_failed, 0)
        self.assertEqual(sim.num_completed, 2)

        # let time pass and update
        time_mock.return_value += 10 * dt
        sim.update()
        self.assertEqual(sim.num_queued, 0)
        self.assertEqual(sim.num_running, 0)
        self.assertEqual(sim.num_failed, 0)
        self.assertEqual(sim.num_completed, 3)

        # test failure rate
        options = BackendSimulatorOptions(max_concurrency=2, failure_rate=1.0)
        sim3 = BackendSimulator(options=options)
        sim3.run_trial(0, dt)
        self.assertEqual(sim3.num_queued, 0)
        self.assertEqual(sim3.num_running, 0)
        self.assertEqual(sim3.num_failed, 1)
        self.assertEqual(sim3.num_completed, 0)

        with self.subTest("Test logging"):
            with self.assertLogs(
                logger=get_logger("utils.testing.backend_simulator"), level="INFO"
            ):
                sim3.update()

            non_verbose_simulator = BackendSimulator(
                options=options, verbose_logging=False
            )
            with self.assertLogs(
                logger=get_logger("utils.testing.backend_simulator"), level="DEBUG"
            ):
                non_verbose_simulator.update()

    def test_backend_simulator_internal_clock(self) -> None:
        sim = get_backend_simulator_with_trials()
        self.assertEqual(len(sim.all_trials), 3)
        self.assertEqual(sim.time, 0.0)
        self.assertEqual(sim.num_queued, 1)
        self.assertEqual(
            sim.lookup_trial_index_status(trial_index=0), TrialStatus.RUNNING
        )
        self.assertEqual(
            sim.lookup_trial_index_status(trial_index=1), TrialStatus.RUNNING
        )
        self.assertEqual(
            sim.lookup_trial_index_status(trial_index=2), TrialStatus.STAGED
        )

        sim.update()
        self.assertEqual(sim.num_completed, 1)
        self.assertEqual(sim.num_running, 2)
        self.assertEqual(
            sim.lookup_trial_index_status(trial_index=1), TrialStatus.COMPLETED
        )

        sim.update()
        self.assertEqual(sim.num_completed, 2)
        self.assertEqual(sim.num_running, 1)
        self.assertEqual(
            sim.lookup_trial_index_status(trial_index=0), TrialStatus.COMPLETED
        )

        sim.stop_trial(trial_index=2)
        sim.update()
        self.assertEqual(
            sim.lookup_trial_index_status(trial_index=2), TrialStatus.COMPLETED
        )
        self.assertEqual(
            # pyre-fixme[16]: Optional type has no attribute `sim_completed_time`.
            sim.get_sim_trial_by_index(trial_index=2).sim_completed_time,
            2.0,
        )
        with self.assertRaisesRegex(ValueError, "Trial 100 not found in simulator"):
            sim.lookup_trial_index_status(trial_index=100)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from logging import Logger
from typing import Mapping, Sequence

from ax.adapter.base import Adapter
from ax.adapter.torch import TorchAdapter
from ax.analysis.plotly.plotly_analysis import PlotlyAnalysisCard
from ax.analysis.plotly.scatter import ScatterPlot
from ax.analysis.utils import extract_relevant_adapter
from ax.core.arm import Arm
from ax.core.experiment import Experiment
from ax.core.optimization_config import MultiObjectiveOptimizationConfig
from ax.core.outcome_constraint import ScalarizedOutcomeConstraint
from ax.core.trial_status import TrialStatus
from ax.exceptions.core import AxError, UnsupportedError, UserInputError
from ax.generation_strategy.generation_strategy import GenerationStrategy
from ax.generators.torch.botorch_modular.generator import BoTorchGenerator
from ax.generators.torch.botorch_modular.multi_acquisition import MultiAcquisition
from ax.utils.common.logger import get_logger
from botorch.acquisition.analytic import LogProbabilityOfFeasibility, PosteriorMean
from pyre_extensions import override

logger: Logger = get_logger(__name__)
OBJ_PFEAS_CARDGROUP_SUBTITLE = (
    "This plot shows <b>newly generated arms</b> with optimal trade-offs between"
    " <b>Ax model-estimated effect on the objective (x-axis)</b> and <b>Ax-model"
    " estimated probability of satisfying the constraints (y-axis)</b>. This plot"
    " is useful for understanding: 1) how tight the constraints are (sometimes the"
    " constraints can be configured too conservatively, making it difficult to find"
    " an arm that improves the objective(s) while satisfying the constraints),"
    " 2) how much headroom there is with the current optimization configuration "
    " (objective(s) and constraints). <b>If arms that are likely feasible (y-axis),"
    " do not improve your objective enough, revisiting your optimization config and"
    " relaxing the constraints may be helpful.</b> This analysis can be computed"
    " adhoc in a notebook environment, and will change with modifications to the"
    " optimization config, so you can understand the potential impact of optimization"
    " config modifications prior to running another iteration. Get in touch with the"
    " Ax developers for pointers on including these arms in a trial or running this"
    " via a notebook."
)


class ObjectivePFeasibleFrontierPlot(ScatterPlot):
    """
    Plotly Scatter plot for the objective vs p(feasible). Each arm is represented by
    a single point with 95% confidence intervals if the data is available. Effects are
    the predicted effects using a model.
    """

    def __init__(
        self,
        relativize: bool = False,
        trial_index: int | None = None,
        trial_statuses: Sequence[TrialStatus] | None = None,
        additional_arms: Sequence[Arm] | None = None,
        labels: Mapping[str, str] | None = None,
        show_pareto_frontier: bool = False,
        num_points_to_generate: int = 100,
    ) -> None:
        """
        Args:
            relativize: Whether to relativize the effects of each arm against the status
                quo arm. If multiple status quo arms are present, relativize each arm
                against the status quo arm from the same trial.
            trial_index: If present, only use arms from the trial with the given index.
            additional_arms: If present, include these arms in the plot in addition to
                the arms in the experiment. These arms will be marked as belonging to a
                trial with index -1.
            labels: A mapping from metric names to labels to use in the plot. If a label
                is not provided for a metric, the metric name will be used.
            show_pareto_frontier: Whether to draw a line representing the Pareto
                frontier for the two metrics on the plot.
            num_points_to_generate: The number of points to generate on the frontier.
                Ideally this should be sufficiently large to provide a frontier with
                reasonably good coverage.
        """

        self.y_metric_name = "p_feasible"
        self.use_model_predictions = True
        self.relativize = relativize
        self.trial_index = trial_index
        self.trial_statuses = trial_statuses
        self._additional_arms: list[Arm] = (
            [] if additional_arms is None else list(additional_arms)
        )
        self.additional_arms = additional_arms
        self.labels: dict[str, str] = {**labels} if labels is not None else {}
        self.labels.setdefault("p_feasible", "% Chance of Satisfying the Constraints")
        self.show_pareto_frontier = show_pareto_frontier
        self.num_points_to_generate = num_points_to_generate
        self.title = (
            f"Modeled {'Relativized ' if self.relativize else ''} Effect on the"
            " Objective vs  % Chance of Satisfying the Constraints"
        )

    @override
    def compute(
        self,
        experiment: Experiment | None = None,
        generation_strategy: GenerationStrategy | None = None,
        adapter: Adapter | None = None,
    ) -> PlotlyAnalysisCard:
        if experiment is None:
            raise UserInputError(
                "ObjectivePFeasibleFrontierPlot requires an Experiment."
            )
        if experiment.optimization_config is None:
            raise UserInputError("Optimization_config must be set to compute frontier.")
        elif isinstance(
            experiment.optimization_config, MultiObjectiveOptimizationConfig
        ):
            raise UnsupportedError("Multi-objective optimization is not supported.")
        elif len(experiment.optimization_config.outcome_constraints) == 0:
            raise UserInputError(
                "Plotting the objective-p(feasible) frontier requires at least one "
                "outcome constraint."
            )
        elif any(
            isinstance(oc, ScalarizedOutcomeConstraint)
            for oc in experiment.optimization_config.outcome_constraints
        ):
            raise UnsupportedError(
                "Scalarized outcome constraints are not supported yet."
            )
        self.x_metric_name = experiment.optimization_config.objective.metric.name
        relevant_adapter = extract_relevant_adapter(
            experiment=experiment,
            generation_strategy=generation_strategy,
            adapter=adapter,
        )

        if not isinstance(relevant_adapter, TorchAdapter) or not isinstance(
            relevant_adapter.generator, BoTorchGenerator
        ):
            raise AxError(
                "The Objective vs P(feasible) plot cannot be computed using the"
                f" current Adapter ({relevant_adapter}) and generator"
                f" ({relevant_adapter.generator}). Only TorchAdapters using"
                " BoTorchGenerators are supported."
            )
        # Generate arms on the objective p_feasible frontier
        # Specify to optimize multiple acquisition functions (via
        # MultiAcquisition). This will optimize the PosteriorMean and the
        # LogProbabilityOfFeasibility using MOO.
        generator = relevant_adapter.generator
        orig_acquisition_class = generator.acquisition_class
        orig_botorch_acqf_classes_with_options = (
            generator._botorch_acqf_classes_with_options
        )
        generator.acquisition_class = MultiAcquisition
        generator._botorch_acqf_classes_with_options = [
            (PosteriorMean, {}),
            (LogProbabilityOfFeasibility, {}),
        ]
        frontier_gr = relevant_adapter.gen(n=self.num_points_to_generate)
        # in case this generator is used again, restore the original settings
        generator.acquisition_class = orig_acquisition_class
        generator._botorch_acqf_classes_with_options = (
            orig_botorch_acqf_classes_with_options
        )
        self.additional_arms = self._additional_arms + [
            Arm(name=f"frontier_{i}", parameters=a.parameters)
            for i, a in enumerate(frontier_gr.arms)
        ]
        return super().compute(experiment=experiment, adapter=relevant_adapter)



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict


from contextlib import nullcontext
from dataclasses import replace
from itertools import product
from unittest.mock import Mock, patch

import numpy as np

import torch
from ax.benchmark.benchmark_runner import _add_noise, BenchmarkRunner
from ax.benchmark.benchmark_test_functions.botorch_test import BoTorchTestFunction
from ax.benchmark.benchmark_test_functions.surrogate import SurrogateTestFunction

from ax.benchmark.benchmark_test_functions.synthetic import IdentityTestFunction
from ax.benchmark.problems.synthetic.hss.jenatton import (
    get_jenatton_benchmark_problem,
    Jenatton,
)
from ax.benchmark.testing.benchmark_stubs import (
    DummyTestFunction,
    get_jenatton_trials,
    get_soo_surrogate_test_function,
)
from ax.core.arm import Arm
from ax.core.batch_trial import BatchTrial
from ax.core.experiment import Experiment
from ax.core.search_space import SearchSpace
from ax.core.trial import Trial
from ax.core.trial_status import TrialStatus
from ax.exceptions.core import UnsupportedError
from ax.utils.common.testutils import TestCase

from botorch.test_functions.synthetic import Ackley, ConstrainedHartmann, Hartmann
from botorch.utils.transforms import normalize
from pandas import DataFrame
from pyre_extensions import assert_is_instance, none_throws


class TestBenchmarkRunner(TestCase):
    def setUp(self) -> None:
        super().setUp()
        self.maxDiff = None

    def test_simulated_backend_runner(self) -> None:
        # Initialize
        runner = BenchmarkRunner(
            test_function=Jenatton(outcome_names=["objective"]),
            step_runtime_function=lambda params: params["x1"] + 1,
            max_concurrency=2,
        )
        simulated_backend_runner = none_throws(runner.simulated_backend_runner)
        simulator = simulated_backend_runner.simulator
        self.assertEqual(simulator.max_concurrency, 2)
        self.assertTrue(simulator.use_internal_clock)
        self.assertEqual(simulator.time, 0)
        trials = get_jenatton_trials(n_trials=3)

        # Run trials in parallel
        runner.run(trial=trials[0])
        runner.run(trial=trials[1])
        self.assertEqual(simulator.time, 0)

        # Check status (updating clock)
        statuses = runner.poll_trial_status(trials=[trials[0], trials[1]])
        self.assertEqual(simulator.time, 1)
        self.assertEqual(statuses[TrialStatus.COMPLETED], {0})
        self.assertEqual(statuses[TrialStatus.RUNNING], {1})

        # Run another trial
        runner.run(trial=trials[2])
        self.assertEqual(simulator.time, 1)
        statuses = runner.poll_trial_status(trials=list(trials.values()))
        self.assertEqual(simulator.time, 2)
        self.assertEqual(statuses[TrialStatus.COMPLETED], {0, 1})
        self.assertEqual(statuses[TrialStatus.RUNNING], {2})

    def test_runner(self) -> None:
        botorch_cases = [
            (
                BoTorchTestFunction(
                    botorch_problem=test_problem_class(dim=6),
                    modified_bounds=modified_bounds,
                    outcome_names=outcome_names,
                ),
                noise_std,
                num_outcomes,
            )
            for (test_problem_class, num_outcomes, outcome_names) in (
                (Hartmann, 1, ["objective_0"]),
                (ConstrainedHartmann, 2, ["objective_0", "constraint"]),
            )
            for modified_bounds, noise_std in product(
                (None, [(0.0, 2.0)] * 6),
                (0.0, [0.1] * num_outcomes),
            )
        ]
        param_based_cases = [
            (
                DummyTestFunction(dim=6, num_outcomes=num_outcomes),
                noise_std,
                num_outcomes,
            )
            for num_outcomes in (1, 2)
            for noise_std in (0.0, [float(i) for i in range(num_outcomes)])
        ]
        surrogate_cases = [
            (get_soo_surrogate_test_function(lazy=False), noise_std, 1)
            for noise_std in (0.0, 1.0, [0.0], [1.0])
        ]
        for test_function, noise_std, num_outcomes in (
            botorch_cases + param_based_cases + surrogate_cases
        ):
            # Set up outcome names
            if isinstance(test_function, BoTorchTestFunction):
                if isinstance(test_function.botorch_problem, ConstrainedHartmann):
                    outcome_names = ["objective_0", "constraint"]
                else:
                    outcome_names = ["objective_0"]
            elif isinstance(test_function, DummyTestFunction):
                outcome_names = [f"objective_{i}" for i in range(num_outcomes)]
            else:  # SurrogateTestFunction
                outcome_names = ["branin"]

            # Set up runner
            runner = BenchmarkRunner(test_function=test_function, noise_std=noise_std)

            test_description = f"{test_function=}, {noise_std=}"
            with self.subTest(
                f"Test basic construction, {test_function=}, {noise_std=}"
            ):
                self.assertIs(runner.test_function, test_function)
                self.assertEqual(runner.outcome_names, outcome_names)
                if isinstance(noise_std, list):
                    self.assertEqual(
                        runner.get_noise_stds(),
                        dict(zip(runner.outcome_names, noise_std)),
                    )
                else:  # float
                    self.assertEqual(
                        runner.get_noise_stds(),
                        {name: noise_std for name in runner.outcome_names},
                    )

                # check equality
                new_runner = replace(
                    runner,
                    test_function=BoTorchTestFunction(
                        botorch_problem=Ackley(),
                        outcome_names=test_function.outcome_names,
                    ),
                )
                self.assertNotEqual(runner, new_runner)

                self.assertEqual(runner, runner)
                if isinstance(test_function, BoTorchTestFunction):
                    self.assertEqual(
                        test_function.botorch_problem.bounds.dtype, torch.double
                    )

            is_botorch = isinstance(test_function, BoTorchTestFunction)
            with self.subTest(f"test `get_Y_true()`, {test_description}"):
                dim = 6 if is_botorch else 9
                X = torch.rand(1, dim, dtype=torch.double)
                param_names = (
                    [f"x{i}" for i in range(6)]
                    if is_botorch
                    else list(
                        get_jenatton_benchmark_problem().search_space.parameters.keys()
                    )
                )
                params = dict(zip(param_names, (x.item() for x in X.unbind(-1))))

                with (
                    nullcontext()
                    if not isinstance(test_function, SurrogateTestFunction)
                    else patch.object(
                        # pyre-fixme: BenchmarkTestFunction` has no attribute
                        # `_surrogate`.
                        runner.test_function._surrogate,
                        "predict",
                        return_value=({"branin": [4.2]}, None),
                    )
                ):
                    Y = runner.get_Y_true(params=params)

                if (
                    isinstance(test_function, BoTorchTestFunction)
                    and test_function.modified_bounds is not None
                ):
                    X_tf = normalize(
                        X,
                        torch.tensor(
                            test_function.modified_bounds, dtype=torch.double
                        ).T,
                    )
                else:
                    X_tf = X
                if isinstance(test_function, BoTorchTestFunction):
                    botorch_problem = test_function.botorch_problem
                    obj = botorch_problem.evaluate_true(X_tf)
                    if isinstance(botorch_problem, ConstrainedHartmann):
                        expected_Y = np.array(
                            [
                                [obj.item()],
                                [botorch_problem.evaluate_slack(X_tf).item()],
                            ]
                        )
                    else:
                        expected_Y = obj.numpy()[:, None]
                        self.assertEqual(expected_Y.ndim, 2)
                elif isinstance(test_function, SurrogateTestFunction):
                    expected_Y = np.array([[4.2]])
                else:
                    expected_Y = np.full((2, 1), X.pow(2).sum().item())
                self.assertTrue(np.allclose(Y, expected_Y))

            with self.subTest(f"test `run()`, {test_description}"):
                trial = Mock(spec=Trial)
                arm = Arm(name="0_0", parameters=params)
                trial.arms = [arm]
                trial.arm = arm
                trial.index = 0

                with (
                    nullcontext()
                    if not isinstance(test_function, SurrogateTestFunction)
                    else patch.object(
                        runner.test_function._surrogate,
                        "predict",
                        return_value=({"branin": [4.2]}, None),
                    )
                ):
                    res = runner.run(trial=trial)["benchmark_metadata"].dfs
                df = next(iter(res.values()))
                self.assertEqual({"0_0"}, set(df["arm_name"]))
                self.assertEqual(
                    set(runner.test_function.outcome_names), set(res.keys())
                )

                for i, df in enumerate(res.values()):
                    if isinstance(noise_std, list):
                        self.assertEqual(df["sem"].item(), noise_std[i])
                        if all(n == 0 for n in noise_std):
                            self.assertTrue(np.array_equal(df["mean"], Y[i, :]))
                    else:  # float
                        self.assertEqual(df["sem"].item(), noise_std)
                        if noise_std == 0:
                            self.assertTrue(np.array_equal(df["mean"], Y[i, :]))

            with self.subTest(f"test `poll_trial_status()`, {test_description}"):
                self.assertEqual(
                    {TrialStatus.COMPLETED: {0}}, runner.poll_trial_status([trial])
                )

            with self.subTest(f"test `serialize_init_args()`, {test_description}"):
                with self.assertRaisesRegex(
                    UnsupportedError, "serialize_init_args is not a supported method"
                ):
                    BenchmarkRunner.serialize_init_args(obj=runner)
                with self.assertRaisesRegex(
                    UnsupportedError, "deserialize_init_args is not a supported method"
                ):
                    BenchmarkRunner.deserialize_init_args({})

    def test__add_noise(self) -> None:
        np.random.seed(0)
        y_true = np.arange(6)
        arm_name = ["0_0", "0_1", "0_0", "0_1", "0_0", "0_1"]
        metric_name = ["foo", "foo", "bar", "bar", "baz", "baz"]

        df = DataFrame(
            {"Y_true": y_true, "metric_name": metric_name, "arm_name": arm_name}
        )

        noise_stds = {"foo": 1, "bar": 2, "baz": 3}
        arm_weights = {"0_0": 1, "0_1": 2}
        result = _add_noise(df=df, noise_stds=noise_stds, arm_weights=arm_weights)
        self.assertEqual(set(result.columns), set(df.columns) | {"mean", "sem"})
        expected_sem = df["metric_name"].map(noise_stds) / np.sqrt(
            df["arm_name"].map(arm_weights) / 3
        )
        self.assertEqual(result["sem"].tolist(), expected_sem.tolist())
        noise = df["mean"] - df["Y_true"]
        self.assertNotEqual(noise.std(), 0)

        z_scores = noise / expected_sem
        self.assertNotEqual(z_scores.std(), 0)

        chi_squared_stat = (z_scores**2).sum()
        # None of these assertions would have failed in 10M simulations.
        # Each has some tolerance from the most extreme value seen in 10M sims.
        self.assertGreater(chi_squared_stat, 0.005)
        self.assertLess(chi_squared_stat, 45)
        self.assertLess(np.abs(z_scores).min(), 2)
        self.assertGreater(z_scores.max(), 0.05)

    def test_get_noise_stds(self) -> None:
        test_function = BoTorchTestFunction(
            botorch_problem=Hartmann(dim=6),
            outcome_names=["objective_0"],
        )
        expected_noise_sd_dict = {"objective_0": 1.0}
        with self.subTest("float noise_std"):
            runner = BenchmarkRunner(test_function=test_function, noise_std=1.0)
            self.assertDictEqual(runner.get_noise_stds(), expected_noise_sd_dict)

        with self.subTest("int noise_std"):
            runner = BenchmarkRunner(test_function=test_function, noise_std=1)
            self.assertDictEqual(runner.get_noise_stds(), expected_noise_sd_dict)

        with self.subTest("dict noise_std"):
            runner = BenchmarkRunner(
                test_function=test_function, noise_std=expected_noise_sd_dict
            )
            self.assertDictEqual(runner.get_noise_stds(), expected_noise_sd_dict)

        with self.subTest("list noise_std"):
            runner = BenchmarkRunner(test_function=test_function, noise_std=[1.0])
            self.assertDictEqual(runner.get_noise_stds(), expected_noise_sd_dict)

    def test_heterogeneous_noise(self) -> None:
        outcome_names = ["objective_0", "constraint"]
        noise_dict = {"objective_0": 0.1, "constraint": 0.05}
        for noise_std in [[0.1, 0.05], noise_dict]:
            runner = BenchmarkRunner(
                test_function=BoTorchTestFunction(
                    botorch_problem=ConstrainedHartmann(dim=6),
                    outcome_names=outcome_names,
                ),
                noise_std=noise_std,
            )
            self.assertDictEqual(
                assert_is_instance(runner.get_noise_stds(), dict), noise_dict
            )

            X = torch.rand(1, 6, dtype=torch.double)
            arm = Arm(
                name="0_0",
                parameters={f"x{i}": x.item() for i, x in enumerate(X.unbind(-1))},
            )
            trial = Mock(spec=Trial)
            trial.arms = [arm]
            trial.arm = arm
            trial.index = 0
            res = runner.run(trial=trial)["benchmark_metadata"].dfs
            self.assertEqual(
                set(outcome_names),
                res.keys(),
            )
            obj_df = res["objective_0"]
            self.assertEqual(len(obj_df), 1)
            self.assertEqual(
                {
                    "arm_name",
                    "metric_name",
                    "mean",
                    "sem",
                    "trial_index",
                    "step",
                    "virtual runtime",
                },
                set(obj_df.columns),
            )
            self.assertEqual(obj_df["arm_name"].item(), "0_0")
            self.assertEqual(obj_df["sem"].item(), 0.1)
            self.assertEqual(res["constraint"]["sem"].item(), 0.05)

            with self.subTest("heterogeneous arm weights"):
                arm_0 = Arm(name="0_0", parameters={f"x{i}": 0.0 for i in range(6)})
                arm_1 = Arm(name="0_1", parameters={f"x{i}": 0.5 for i in range(6)})
                trial = Mock(spec=BatchTrial)
                trial.arms = [arm_0, arm_1]
                trial.index = 0
                # arm stds get multiplied by sqrt([25/9, 25/16])
                trial.arm_weights = {arm_0: 9, arm_1: 16}
                res = runner.run(trial=trial)["benchmark_metadata"].dfs
                expected_relative_noise_levels = np.array([5 / 3, 5 / 4])
                for metric_name, df in res.items():
                    self.assertTrue(
                        np.array_equal(
                            df["sem"],
                            noise_dict[metric_name] * expected_relative_noise_levels,
                        )
                    )

    def test_with_learning_curve(self) -> None:
        test_function = IdentityTestFunction(outcome_names=["foo", "bar"], n_steps=10)

        params = {"x0": 1.2}
        runner = BenchmarkRunner(test_function=test_function, noise_std=0.0)

        experiment = Experiment(
            name="test",
            is_test=True,
            runner=runner,
            search_space=Mock(spec=SearchSpace),
        )

        for noise_std in [0.0, 0.1]:
            with self.subTest(noise_std=noise_std):
                runner = BenchmarkRunner(
                    test_function=test_function, noise_std=noise_std
                )

                trial = Trial(experiment=experiment)
                arm = Arm(name="0_0", parameters=params)
                trial.add_arm(arm=arm)
                metadata_dict = runner.run(trial=trial)
                self.assertEqual({"benchmark_metadata"}, metadata_dict.keys())
                metadata = metadata_dict["benchmark_metadata"].dfs
                self.assertEqual(set(metadata.keys()), set(test_function.outcome_names))
                for df in metadata.values():
                    self.assertEqual(len(df), 10)
                    self.assertTrue((df["arm_name"] == "0_0").all())
                    self.assertTrue(np.array_equal(df["step"], np.arange(10)))
                    self.assertTrue((df["sem"] == noise_std).all())

                noiseless = test_function.evaluate_true(params=params)
                all_close = np.allclose(metadata["foo"]["mean"], noiseless[0, :])
                self.assertEqual(all_close, noise_std == 0.0)

        with self.subTest("with SimulatedBackendRunner"):
            runner = BenchmarkRunner(
                test_function=test_function, noise_std=0.0, max_concurrency=2
            )

            arm = Arm(name="0_0", parameters=params)
            trial = Trial(experiment=experiment)
            trial.add_arm(arm=arm)
            metadata = runner.run(trial=trial)["benchmark_metadata"]
            backend_simulator = none_throws(metadata.backend_simulator)
            self.assertNotIn(trial.index, backend_simulator._completed)
            sim_trial = none_throws(
                backend_simulator.get_sim_trial_by_index(trial.index)
            )
            self.assertIsNone(sim_trial.sim_completed_time)
            self.assertEqual(sim_trial.sim_start_time, 0)
            self.assertEqual(backend_simulator.time, 0)

    def test_heterogeneous_step_runtime(self) -> None:
        n_steps = 10
        test_function = IdentityTestFunction(
            outcome_names=["foo", "bar"], n_steps=n_steps
        )
        runner = BenchmarkRunner(
            test_function=test_function,
            noise_std=0.0,
            step_runtime_function=lambda params: params["x0"],
        )
        experiment = Experiment(
            name="test",
            is_test=True,
            runner=runner,
            search_space=Mock(spec=SearchSpace),
        )
        trial = BatchTrial(experiment=experiment)
        arm_0_step_time = 0.5
        arm_1_step_time = 1.5
        trial.add_arm(Arm(name="0_0", parameters={"x0": arm_0_step_time}))
        trial.add_arm(Arm(name="0_1", parameters={"x0": arm_1_step_time}))
        df = runner.run(trial=trial)["benchmark_metadata"].dfs["foo"]
        total_runtime = df.groupby("arm_name")["virtual runtime"].max()
        self.assertEqual(
            total_runtime.to_dict(),
            {"0_0": arm_0_step_time * n_steps, "0_1": arm_1_step_time * n_steps},
        )
        max_step = df.groupby("arm_name")["step"].max()
        self.assertEqual(max_step.to_list(), [9, 9])

        with self.subTest("Test runtimes non-negative"):
            trial = BatchTrial(experiment=experiment)
            trial.add_arm(Arm(name="0_0", parameters={"x0": -1}))
            with self.assertRaisesRegex(
                ValueError, "Step duration must be non-negative"
            ):
                runner.run(trial=trial)

    def test_wrong_noise_std_keys(self) -> None:
        test_function = IdentityTestFunction(outcome_names=["foo", "bar"])
        runner = BenchmarkRunner(test_function=test_function, noise_std={"alpaca": 4})
        with self.assertRaisesRegex(
            ValueError, "Noise std must have keys equal to outcome names"
        ):
            runner.get_noise_stds()



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from typing import Optional, TYPE_CHECKING

from ax.adapter.data_utils import ExperimentData
from ax.adapter.transforms.base import Transform
from ax.core.arm import Arm
from ax.core.observation import Observation, ObservationFeatures
from ax.core.parameter import ChoiceParameter, FixedParameter, ParameterType
from ax.core.search_space import RobustSearchSpace, SearchSpace
from ax.core.types import TParameterization, TParamValue
from ax.exceptions.core import UnsupportedError
from ax.generators.types import TConfig
from pyre_extensions import assert_is_instance, none_throws

if TYPE_CHECKING:
    # import as module to make sphinx-autodoc-typehints happy
    from ax import adapter as adapter_module  # noqa F401


class SearchSpaceToChoice(Transform):
    """Replaces the search space with a single choice parameter, whose values
    are the signatures of the arms observed in the data.

    This transform is meant to be used with ThompsonSampler.

    Choice parameter will be unordered unless config["use_ordered"] specifies
    otherwise.

    Transform is done in-place.
    """

    requires_data_for_initialization: bool = True

    def __init__(
        self,
        search_space: SearchSpace | None = None,
        observations: list[Observation] | None = None,
        experiment_data: ExperimentData | None = None,
        adapter: Optional["adapter_module.base.Adapter"] = None,
        config: TConfig | None = None,
    ) -> None:
        assert search_space is not None, "SearchSpaceToChoice requires search space"
        super().__init__(
            search_space=search_space,
            observations=observations,
            experiment_data=experiment_data,
            adapter=adapter,
            config=config,
        )
        if any(p.is_fidelity for p in search_space.parameters.values()):
            raise ValueError(
                "Cannot perform SearchSpaceToChoice conversion if fidelity "
                "parameters are present"
            )
        if isinstance(search_space, RobustSearchSpace):
            raise UnsupportedError(
                "SearchSpaceToChoice transform is not supported for RobustSearchSpace."
            )
        self.parameter_name = "arms"
        self.parameter_names: list[str] = list(search_space.parameters)
        if experiment_data is not None:
            arm_data = experiment_data.arm_data[self.parameter_names]
            arm_data = arm_data[self.parameter_names]
            self.signature_to_parameterization: dict[str, TParameterization] = {
                Arm(parameters=row).signature: row.copy()
                for row in arm_data.to_dict(orient="records")
            }
        else:
            self.signature_to_parameterization: dict[str, TParameterization] = {
                Arm(
                    parameters=obs.features.parameters.copy()
                ).signature: obs.features.parameters
                for obs in none_throws(observations)
            }

    def _transform_search_space(self, search_space: SearchSpace) -> SearchSpace:
        values: list[TParamValue] = list(self.signature_to_parameterization.keys())
        if len(values) > 1:
            parameter = ChoiceParameter(
                name=self.parameter_name,
                parameter_type=ParameterType.STRING,
                values=values,
                is_ordered=assert_is_instance(
                    self.config.get("use_ordered", False),
                    bool,
                ),
                sort_values=False,
            )
        else:
            parameter = FixedParameter(
                name=self.parameter_name,
                parameter_type=ParameterType.STRING,
                value=values[0],
            )
        return SearchSpace(parameters=[parameter])

    def transform_observation_features(
        self, observation_features: list[ObservationFeatures]
    ) -> list[ObservationFeatures]:
        for obsf in observation_features:
            # if obsf.parameters is not an empty dict
            if len(obsf.parameters) != 0:
                obsf.parameters = {
                    self.parameter_name: Arm(parameters=obsf.parameters).signature
                }
        return observation_features

    def untransform_observation_features(
        self, observation_features: list[ObservationFeatures]
    ) -> list[ObservationFeatures]:
        for obsf in observation_features:
            # Do not untransform empty dict as it wasn't transformed in the first place
            if len(obsf.parameters) != 0:
                signature = assert_is_instance(
                    obsf.parameters[self.parameter_name], str
                )
                obsf.parameters = self.signature_to_parameterization[signature]
        return observation_features

    def transform_experiment_data(
        self, experiment_data: ExperimentData
    ) -> ExperimentData:
        arm_data = experiment_data.arm_data
        if arm_data.empty:
            return experiment_data
        arm_data[self.parameter_name] = arm_data[self.parameter_names].apply(
            lambda row: Arm(parameters=row).signature, axis=1
        )
        arm_data = arm_data[[self.parameter_name, "metadata"]]
        return ExperimentData(
            arm_data=arm_data,
            observation_data=experiment_data.observation_data,
        )



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-safe

import plotly.express as px

# Colors sampled from Botorch flame logo
BOTORCH_COLOR_SCALE = [
    "#f7931e",  # Botorch orange
    "#eb882d",
    "#df7d3c",
    "#d3724b",
    "#c7685a",
    "#bb5d69",
    "#af5278",
    "#a34887",
    "#973d96",
    "#8b32a5",
    "#7f28b5",  # Botorch purple
    "#792fbb",
    "#7436c1",
    "#6f3dc7",
    "#6a44cd",
    "#654bd4",
    "#6052da",
    "#5b59e0",
    "#5660e6",
    "#5167ec",
    "#4c6ef3",  # Botorch blue
]
AX_BLUE = "#5078f9"  # rgb code: rgb(80, 120, 249)
LIGHT_AX_BLUE = "#adc0fd"  # rgb(173, 192, 253)

CONSTRAINT_VIOLATION_RED = "red"

# Use the same continuous sequential color scale for all plots. PRGn uses purples for
# low values and transitions to greens for high values.
METRIC_CONTINUOUS_COLOR_SCALE: list[str] = px.colors.diverging.Earth
COLOR_FOR_INCREASES: str = METRIC_CONTINUOUS_COLOR_SCALE[5]  # blue
COLOR_FOR_DECREASES: str = METRIC_CONTINUOUS_COLOR_SCALE[2]  # brown

DISCRETE_ARM_SCALE = px.colors.qualitative.Alphabet



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from collections.abc import Iterable, Mapping, Sequence
from dataclasses import dataclass, field
from math import sqrt
from typing import Any

import numpy as np
import numpy.typing as npt
import pandas as pd
from ax.benchmark.benchmark_step_runtime_function import TBenchmarkStepRuntimeFunction

from ax.benchmark.benchmark_test_function import BenchmarkTestFunction
from ax.benchmark.benchmark_trial_metadata import BenchmarkTrialMetadata
from ax.core.base_trial import BaseTrial, TrialStatus
from ax.core.batch_trial import BatchTrial
from ax.core.runner import Runner
from ax.core.types import TParamValue
from ax.exceptions.core import UnsupportedError
from ax.runners.simulated_backend import SimulatedBackendRunner
from ax.utils.common.serialization import TClassDecoderRegistry, TDecoderRegistry
from ax.utils.testing.backend_simulator import BackendSimulator, BackendSimulatorOptions
from pyre_extensions import assert_is_instance


def _dict_of_arrays_to_df(
    Y_true_by_arm: Mapping[str, npt.NDArray],
    step_duration_by_arm: Mapping[str, float],
    outcome_names: Sequence[str],
) -> pd.DataFrame:
    """
    Return a DataFrame with columns
    ["metric_name", "arm_name", "Y_true", "step", and "virtual runtime"].

    When the trial produces MapData, the "step" column is 0, 1, 2, ...., and
    "virtual runtime" contains cumulative time for each element of the
    progression. When the trial does not produce MapData, the "step" column is
    just 0, and "virtual runtime" is the total runtime of the trial.

    Args:
        Y_true_by_arm: A mapping from arm name to a 2D arrays each with shape
            (len(outcome_names), n_steps).
        step_duration_by_arm: A mapping from arm name to a number representing
            the runtime of each step.
        outcome_names: The names of the outcomes; will be mapped to the first
            dimension of each array in ``Y_true_by_arm``.
    """
    df = pd.concat(
        [
            pd.DataFrame(
                {
                    "metric_name": outcome_name,
                    "arm_name": arm_name,
                    "Y_true": y_true[i, :],
                    "step": np.arange(y_true.shape[1], dtype=int),
                    "virtual runtime": np.arange(1, y_true.shape[1] + 1, dtype=int)
                    * step_duration_by_arm[arm_name],
                }
            )
            for i, outcome_name in enumerate(outcome_names)
            for arm_name, y_true in Y_true_by_arm.items()
        ],
        ignore_index=True,
    )
    return df


def _add_noise(
    df: pd.DataFrame,
    noise_stds: Mapping[str, float],
    arm_weights: Mapping[str, float] | None,
) -> pd.DataFrame:
    """
    For each ``Y_true`` in ``df``, with metric name ``metric_name`` and
    arm name ``arm_name``, add noise with standard deviation
    ``noise_stds[metric_name] / sqrt_nlzd_arm_weights[arm_name]``,
    where ``sqrt_nlzd_arm_weights = sqrt(arm_weights[arm_name] /
    sum(arm_weights.values())])``.

    Args:
        df: A DataFrame with columns including
            ["metric_name", "arm_name", "Y_true"].
        noise_stds: A mapping from metric name to what the standard
            deviation would be if one arm received the entire
            sample budget.
        arm_weights: Either ``None`` if there is only one ``Arm``, or a
            mapping from ``Arm`` name to the arm's allocation. Using arm
            weights will increase noise levels, since each ``Arm`` is
            assumed to receive a fraction of the total sample budget.

    Returns:
        The original ``df``, now with additional columns ["mean", "sem"].
    """
    noiseless = all(v == 0 for v in noise_stds.values())
    if not noiseless:
        noise_std_ser = df["metric_name"].map(noise_stds)
        if arm_weights is not None:
            nlzd_arm_weights_sqrt = {
                arm_name: sqrt(weight / sum(arm_weights.values()))
                for arm_name, weight in arm_weights.items()
            }
            arm_weights_ser = df["arm_name"].map(nlzd_arm_weights_sqrt)
            df["sem"] = noise_std_ser / arm_weights_ser

        else:
            df["sem"] = noise_std_ser

        df["mean"] = df["Y_true"] + np.random.normal(loc=0, scale=df["sem"])

    else:
        df["sem"] = 0.0
        df["mean"] = df["Y_true"]
    return df


def get_total_runtime(
    trial: BaseTrial,
    step_runtime_function: TBenchmarkStepRuntimeFunction | None,
    n_steps: int,
) -> float:
    """Get the total runtime of a trial."""
    # By default, each step takes 1 virtual second.
    if step_runtime_function is not None:
        max_step_runtime = max(
            step_runtime_function(arm.parameters) for arm in trial.arms
        )
    else:
        max_step_runtime = 1
    return n_steps * max_step_runtime


@dataclass(kw_only=True)
class BenchmarkRunner(Runner):
    """
    A Runner that produces both observed and ground-truth values.

    Observed values equal ground-truth values plus noise, with the noise added
    according to the standard deviations returned by `get_noise_stds()`.

    This runner does require that every benchmark has a ground truth, which
    won't necessarily be true for real-world problems. Such problems fall into
    two categories:
        - If they are deterministic, they can be used with this runner by
          viewing them as noiseless problems where the observed values are the
          ground truth. The observed values will be used for tracking the
          progress of optimization.
        - If they are not deterministic, they are not supported. It is not
          conceptually clear how to benchmark such problems, so we decided to
          not over-engineer for that before such a use case arrives.

    If ``max_concurrency`` is left as default (1), ``step_runtime_function`` is
    None, and ``force_use_simulated_backend`` is False, then trials run serially
    and complete immediately. Otherwise, a ``SimulatedBackendRunner`` is
    constructed to track the status of trials and the (virtual) times at which
    they start and stop.

    Args:
        test_function: A ``BenchmarkTestFunction`` from which to generate
            deterministic data before adding noise.
        noise_std: The standard deviation of the noise added to the data. Can be
            a list or dict to be per-metric.
        step_runtime_function: A function that takes in parameters
            (in ``TParameterization`` format) and returns the runtime of a step.
        max_concurrency: The maximum number of trials that can be running at a
            given time. Typically, this is ``max_pending_trials`` from the
            ``orchestrator_options`` on the ``BenchmarkMethod``.
        force_use_simulated_backend: If True, use the simulated backend even if
            ``max_concurrency`` is 1 and ``step_runtime_function`` is None. This
            is recommended when used with a ``BenchmarkMethod`` that does early
            stopping.
    """

    test_function: BenchmarkTestFunction
    noise_std: float | Sequence[float] | Mapping[str, float] = 0.0
    step_runtime_function: TBenchmarkStepRuntimeFunction | None = None
    max_concurrency: int = 1
    force_use_simulated_backend: bool = False
    simulated_backend_runner: SimulatedBackendRunner | None = field(init=False)

    def __post_init__(self) -> None:
        use_simulated_backend = (
            (self.max_concurrency > 1)
            or (self.step_runtime_function is not None)
            or self.force_use_simulated_backend
        )
        if use_simulated_backend:
            simulator = BackendSimulator(
                options=BackendSimulatorOptions(
                    max_concurrency=self.max_concurrency,
                    # Always use virtual rather than real time for benchmarking
                    internal_clock=0,
                    use_update_as_start_time=True,
                ),
                verbose_logging=False,
            )
            self.simulated_backend_runner = SimulatedBackendRunner(
                simulator=simulator,
                sample_runtime_func=lambda trial: get_total_runtime(
                    trial=trial,
                    step_runtime_function=self.step_runtime_function,
                    n_steps=self.test_function.n_steps,
                ),
            )
        else:
            self.simulated_backend_runner = None

    @property
    def outcome_names(self) -> Sequence[str]:
        """The names of the outcomes."""
        return self.test_function.outcome_names

    def get_Y_true(self, params: Mapping[str, TParamValue]) -> npt.NDArray:
        """Evaluates the test problem.

        Returns:
            An array of ground truth (noiseless) evaluations, with shape
            (len(outcome_names), n_intervals) if is_map is True, and
            (len(outcome_names), 1) otherwise.
        """
        result = np.atleast_1d(self.test_function.evaluate_true(params=params).numpy())
        if result.ndim == 1:
            return result[:, None]
        return result

    def get_noise_stds(self) -> dict[str, float]:
        noise_std = self.noise_std
        if isinstance(noise_std, float | int):
            return {name: float(noise_std) for name in self.outcome_names}
        elif isinstance(noise_std, dict):
            if not set(noise_std.keys()) == set(self.outcome_names):
                raise ValueError(
                    "Noise std must have keys equal to outcome names if given as "
                    "a dict."
                )
            return noise_std
        # list of floats
        return dict(
            zip(self.outcome_names, assert_is_instance(noise_std, list), strict=True)
        )

    def run(self, trial: BaseTrial) -> dict[str, BenchmarkTrialMetadata]:
        """Run the trial by evaluating its parameterization(s).

        Args:
            trial: The trial to evaluate.

        Returns:
            A dictionary {"benchmark_metadata": metadata}, where ``metadata`` is
            a ``BenchmarkTrialMetadata``.
        """
        Y_true_by_arm = {
            arm.name: self.get_Y_true(arm.parameters) for arm in trial.arms
        }

        step_duration_by_arm = {
            arm.name: 1
            if self.step_runtime_function is None
            else self.step_runtime_function(arm.parameters)
            for arm in trial.arms
        }
        for arm_name, duration in step_duration_by_arm.items():
            if duration < 0:
                raise ValueError(
                    "Step duration must be non-negative for each arm. For arm "
                    f"{arm_name}, duration is {duration}."
                )

        df = _dict_of_arrays_to_df(
            Y_true_by_arm=Y_true_by_arm,
            step_duration_by_arm=step_duration_by_arm,
            outcome_names=self.outcome_names,
        )

        arm_weights = (
            {arm.name: w for arm, w in trial.arm_weights.items()}
            if isinstance(trial, BatchTrial)
            else None
        )

        df = _add_noise(
            df=df, noise_stds=self.get_noise_stds(), arm_weights=arm_weights
        )
        df["trial_index"] = trial.index
        df.drop(columns=["Y_true"], inplace=True)

        if self.simulated_backend_runner is not None:
            self.simulated_backend_runner.run(trial=trial)

        dfs = {
            outcome_name: df[df["metric_name"] == outcome_name]
            for outcome_name in self.outcome_names
        }

        metadata = BenchmarkTrialMetadata(
            dfs=dfs,
            backend_simulator=None
            if self.simulated_backend_runner is None
            else self.simulated_backend_runner.simulator,
        )
        return {"benchmark_metadata": metadata}

    def poll_trial_status(
        self, trials: Iterable[BaseTrial]
    ) -> dict[TrialStatus, set[int]]:
        if self.simulated_backend_runner is None:
            return {TrialStatus.COMPLETED: {t.index for t in trials}}
        return self.simulated_backend_runner.poll_trial_status(trials=trials)

    @classmethod
    def serialize_init_args(cls, obj: Any) -> dict[str, Any]:
        """
        It is tricky to use SerializationMixin with instances that have Ax
        objects as attributes, as BenchmarkRunners do. Therefore, serialization
        is not supported.
        """
        raise UnsupportedError(
            "serialize_init_args is not a supported method for BenchmarkRunners."
        )

    @classmethod
    def deserialize_init_args(
        cls,
        args: dict[str, Any],
        decoder_registry: TDecoderRegistry | None = None,
        class_decoder_registry: TClassDecoderRegistry | None = None,
    ) -> dict[str, Any]:
        """
        It is tricky to use SerializationMixin with instances that have Ax
        objects as attributes, as BenchmarkRunners do. Therefore, serialization
        is not supported.
        """
        raise UnsupportedError(
            "deserialize_init_args is not a supported method for BenchmarkRunners."
        )

    def stop(self, trial: BaseTrial, reason: str | None = None) -> dict[str, Any]:
        if self.simulated_backend_runner is None:
            raise UnsupportedError(
                "stop() is not supported for a `BenchmarkRunner` without a "
                "`simulated_backend_runner`, because trials complete "
                "immediately."
            )
        return self.simulated_backend_runner.stop(trial=trial, reason=reason)



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from unittest.mock import MagicMock, patch

import torch
from ax.adapter.factory import get_sobol
from ax.adapter.registry import Generators
from ax.core.observation import ObservationFeatures
from ax.core.trial_status import TrialStatus
from ax.exceptions.core import UserInputError
from ax.exceptions.model import ModelError
from ax.generation_strategy.best_model_selector import (
    ReductionCriterion,
    SingleDiagnosticBestModelSelector,
)
from ax.generation_strategy.generation_node import (
    GenerationNode,
    GenerationStep,
    logger,
    MISSING_MODEL_SELECTOR_MESSAGE,
)
from ax.generation_strategy.generation_node_input_constructors import (
    InputConstructorPurpose,
    NodeInputConstructors,
)
from ax.generation_strategy.generator_spec import GeneratorSpec
from ax.generation_strategy.transition_criterion import MinTrials
from ax.utils.common.constants import Keys
from ax.utils.common.testutils import TestCase
from ax.utils.testing.core_stubs import get_branin_experiment
from ax.utils.testing.mock import mock_botorch_optimize
from botorch.sampling.normal import SobolQMCNormalSampler
from pyre_extensions import none_throws


class TestGenerationNode(TestCase):
    def setUp(self) -> None:
        super().setUp()
        self.sobol_generator_spec = GeneratorSpec(
            generator_enum=Generators.SOBOL,
            model_kwargs={"init_position": 3},
            model_gen_kwargs={"some_gen_kwarg": "some_value"},
        )
        self.mbm_generator_spec = GeneratorSpec(
            generator_enum=Generators.BOTORCH_MODULAR,
            model_kwargs={},
            model_gen_kwargs={},
        )
        self.sobol_generation_node = GenerationNode(
            node_name="test", generator_specs=[self.sobol_generator_spec]
        )
        self.branin_experiment = get_branin_experiment(with_completed_trial=True)
        self.branin_data = self.branin_experiment.lookup_data()
        self.node_short = GenerationNode(
            node_name="test",
            generator_specs=[self.sobol_generator_spec],
            trial_type=Keys.SHORT_RUN,
        )

    def test_init(self) -> None:
        self.assertEqual(
            self.sobol_generation_node.generator_specs, [self.sobol_generator_spec]
        )
        with self.assertRaisesRegex(UserInputError, "Model keys must be unique"):
            GenerationNode(
                node_name="test",
                generator_specs=[self.sobol_generator_spec, self.sobol_generator_spec],
            )
        mbm_specs = [
            GeneratorSpec(generator_enum=Generators.BOTORCH_MODULAR),
            GeneratorSpec(
                generator_enum=Generators.BOTORCH_MODULAR, model_key_override="MBM v2"
            ),
        ]
        with self.assertRaisesRegex(UserInputError, MISSING_MODEL_SELECTOR_MESSAGE):
            GenerationNode(
                node_name="test",
                generator_specs=mbm_specs,
            )
        model_selector = SingleDiagnosticBestModelSelector(
            diagnostic="Fisher exact test p",
            metric_aggregation=ReductionCriterion.MEAN,
            criterion=ReductionCriterion.MIN,
        )
        node = GenerationNode(
            node_name="test",
            generator_specs=mbm_specs,
            best_model_selector=model_selector,
        )
        self.assertEqual(node.generator_specs, mbm_specs)
        self.assertIs(node.best_model_selector, model_selector)

    def test_input_constructor_none(self) -> None:
        self.assertEqual(self.sobol_generation_node._input_constructors, {})
        self.assertEqual(self.sobol_generation_node.input_constructors, {})

    def test_incorrect_trial_type(self) -> None:
        with self.assertRaisesRegex(NotImplementedError, "Trial type must be either"):
            GenerationNode(
                node_name="test",
                generator_specs=[self.sobol_generator_spec],
                trial_type="foo",
            )

    def test_init_with_trial_type(self) -> None:
        node_long = GenerationNode(
            node_name="test",
            generator_specs=[self.sobol_generator_spec],
            trial_type=Keys.LONG_RUN,
        )
        node_default = GenerationNode(
            node_name="test",
            generator_specs=[self.sobol_generator_spec],
        )
        self.assertEqual(self.node_short._trial_type, Keys.SHORT_RUN)
        self.assertEqual(node_long._trial_type, Keys.LONG_RUN)
        self.assertIsNone(node_default._trial_type)

    def test_input_constructor(self) -> None:
        node = GenerationNode(
            node_name="test",
            generator_specs=[self.sobol_generator_spec],
            input_constructors={InputConstructorPurpose.N: NodeInputConstructors.ALL_N},
        )
        self.assertEqual(
            node.input_constructors,
            {InputConstructorPurpose.N: NodeInputConstructors.ALL_N},
        )
        self.assertEqual(
            node._input_constructors,
            {InputConstructorPurpose.N: NodeInputConstructors.ALL_N},
        )

    def test_fit(self) -> None:
        with patch.object(
            self.sobol_generator_spec, "fit", wraps=self.sobol_generator_spec.fit
        ) as mock_generator_spec_fit:
            self.sobol_generation_node._fit(
                experiment=self.branin_experiment,
                data=self.branin_data,
            )
        mock_generator_spec_fit.assert_called_with(
            experiment=self.branin_experiment, data=self.branin_data
        )

    def test_gen(self) -> None:
        with patch.object(
            self.sobol_generator_spec, "gen", wraps=self.sobol_generator_spec.gen
        ) as mock_generator_spec_gen, patch.object(
            self.sobol_generator_spec, "fit", wraps=self.sobol_generator_spec.fit
        ) as mock_generator_spec_fit:
            gr = self.sobol_generation_node.gen(
                experiment=self.branin_experiment,
                data=self.branin_experiment.lookup_data(),
                n=1,
                pending_observations={"branin": []},
            )
            self.assertIsNotNone(gr)
            self.assertEqual(gr._model_key, self.sobol_generator_spec.model_key)
            model_kwargs = gr._model_kwargs
            self.assertIsNotNone(model_kwargs)
            self.assertEqual(model_kwargs.get("init_position"), 3)
        mock_generator_spec_fit.assert_called_with(
            experiment=self.branin_experiment, data=self.branin_experiment.lookup_data()
        )
        mock_generator_spec_gen.assert_called_with(
            experiment=self.branin_experiment,
            data=self.branin_experiment.lookup_data(),
            n=1,
            pending_observations={"branin": []},
            fixed_features=None,
        )

    @mock_botorch_optimize
    def test_gen_with_trial_type(self) -> None:
        mbm_short = GenerationNode(
            node_name="test",
            generator_specs=[
                GeneratorSpec(
                    generator_enum=Generators.BOTORCH_MODULAR,
                    model_kwargs={},
                    model_gen_kwargs={
                        "n": 1,
                        "fixed_features": ObservationFeatures(
                            parameters={},
                            trial_index=0,
                        ),
                    },
                ),
            ],
            trial_type=Keys.SHORT_RUN,
        )
        gr = mbm_short.gen(
            experiment=self.branin_experiment,
            data=self.branin_experiment.lookup_data(),
            pending_observations=None,
            n=2,
        )
        self.assertIsNotNone(gr)
        gen_metadata = gr.gen_metadata
        self.assertIsNotNone(gen_metadata)
        self.assertEqual(gen_metadata["trial_type"], Keys.SHORT_RUN)
        # validate that other fields in gen_metadata are preserved
        self.assertIsNotNone(gen_metadata[Keys.EXPECTED_ACQF_VAL])

    def test_gen_with_no_trial_type(self) -> None:
        gr = self.sobol_generation_node.gen(
            experiment=self.branin_experiment,
            data=self.branin_experiment.lookup_data(),
            pending_observations=None,
            n=2,
        )
        self.assertIsNotNone(gr)
        self.assertNotIn("trial_type", none_throws(gr.gen_metadata))

    @mock_botorch_optimize
    def test_model_gen_kwargs_deepcopy(self) -> None:
        sampler = SobolQMCNormalSampler(torch.Size([1]))
        node = GenerationNode(
            node_name="test",
            generator_specs=[
                GeneratorSpec(
                    generator_enum=Generators.BOTORCH_MODULAR,
                    model_kwargs={},
                    model_gen_kwargs={
                        "n": 1,
                        "fixed_features": ObservationFeatures(
                            parameters={},
                            trial_index=0,
                        ),
                        "model_gen_options": {Keys.ACQF_KWARGS: {"sampler": sampler}},
                    },
                ),
            ],
        )
        dat = self.branin_experiment.lookup_data()
        node.gen(
            experiment=self.branin_experiment,
            data=dat,
            n=1,
            pending_observations={"branin": []},
        )
        # verify that sampler is not modified in-place by checking base samples
        self.assertIs(
            node.generator_spec_to_gen_from.model_gen_kwargs["model_gen_options"][
                Keys.ACQF_KWARGS
            ]["sampler"],
            sampler,
        )
        self.assertIsNone(sampler.base_samples)

    @mock_botorch_optimize
    def test_properties(self) -> None:
        node = GenerationNode(
            node_name="test",
            generator_specs=[
                GeneratorSpec(
                    generator_enum=Generators.BOTORCH_MODULAR,
                    model_kwargs={},
                    model_gen_kwargs={
                        "n": 1,
                        "fixed_features": ObservationFeatures(
                            parameters={},
                            trial_index=0,
                        ),
                    },
                ),
            ],
        )
        self.assertEqual(node.model_to_gen_from_name, "BoTorch")
        node._fit(
            experiment=self.branin_experiment,
            data=self.branin_data,
        )
        self.assertEqual(
            node.generator_spec_to_gen_from.generator_enum,
            node.generator_specs[0].generator_enum,
        )
        self.assertEqual(
            node.generator_spec_to_gen_from.model_kwargs,
            node.generator_specs[0].model_kwargs,
        )
        self.assertEqual(node.model_to_gen_from_name, "BoTorch")
        self.assertEqual(
            node.generator_spec_to_gen_from.model_gen_kwargs,
            node.generator_specs[0].model_gen_kwargs,
        )
        self.assertEqual(
            node.generator_spec_to_gen_from.model_cv_kwargs,
            node.generator_specs[0].model_cv_kwargs,
        )
        self.assertEqual(
            node.generator_spec_to_gen_from.fixed_features,
            node.generator_specs[0].fixed_features,
        )
        self.assertEqual(
            node.generator_spec_to_gen_from.cv_results,
            node.generator_specs[0].cv_results,
        )
        self.assertEqual(
            node.generator_spec_to_gen_from.diagnostics,
            node.generator_specs[0].diagnostics,
        )
        self.assertEqual(node.node_name, "test")
        self.assertEqual(node._unique_id, "test")

    def test_node_string_representation(self) -> None:
        node = GenerationNode(
            node_name="test",
            generator_specs=[
                self.mbm_generator_spec,
            ],
            transition_criteria=[
                MinTrials(threshold=5, only_in_statuses=[TrialStatus.RUNNING])
            ],
        )
        string_rep = str(node)

        self.assertEqual(
            string_rep,
            "GenerationNode(node_name='test', "
            "generator_specs=[GeneratorSpec(generator_enum=BoTorch, "
            "model_key_override=None)], "
            "transition_criteria=[MinTrials(transition_to='None')])",
        )

    def test_single_fixed_features(self) -> None:
        node = GenerationNode(
            node_name="test",
            generator_specs=[
                GeneratorSpec(
                    generator_enum=Generators.BOTORCH_MODULAR,
                    model_kwargs={},
                    model_gen_kwargs={
                        "n": 2,
                        "fixed_features": ObservationFeatures(parameters={"x": 0}),
                    },
                ),
            ],
        )
        self.assertEqual(
            node.generator_spec_to_gen_from.fixed_features,
            ObservationFeatures(parameters={"x": 0}),
        )


class TestGenerationStep(TestCase):
    def setUp(self) -> None:
        super().setUp()
        self.model_kwargs = {"init_position": 5}
        self.sobol_generation_step = GenerationStep(
            generator=Generators.SOBOL,
            num_trials=5,
            model_kwargs=self.model_kwargs,
        )
        self.generator_spec = GeneratorSpec(
            generator_enum=self.sobol_generation_step.generator,
            model_kwargs=self.model_kwargs,
        )

    def test_init(self) -> None:
        self.assertEqual(
            self.sobol_generation_step.generator_specs,
            [self.generator_spec],
        )
        self.assertEqual(self.sobol_generation_step.generator_name, "Sobol")

        named_generation_step = GenerationStep(
            generator=Generators.SOBOL,
            num_trials=5,
            model_kwargs=self.model_kwargs,
            generator_name="Custom Sobol",
        )
        self.assertEqual(named_generation_step.generator_name, "Custom Sobol")

    def test_min_trials_observed(self) -> None:
        with self.assertRaisesRegex(UserInputError, "min_trials_observed > num_trials"):
            GenerationStep(
                generator=Generators.SOBOL,
                num_trials=5,
                min_trials_observed=10,
                model_kwargs=self.model_kwargs,
            )

    def test_init_factory_function(self) -> None:
        with self.assertRaisesRegex(
            UserInputError, "must be a `GeneratorRegistryBase`"
        ):
            # pyre-ignore [6]: Testing deprecated input.
            GenerationStep(generator=get_sobol, num_trials=-1)

    def test_properties(self) -> None:
        step = self.sobol_generation_step
        self.assertEqual(step.generator_spec, self.generator_spec)
        self.assertEqual(step._unique_id, "-1")
        # Make sure that model_kwargs and model_gen_kwargs are synchronized
        # to the underlying model spec.
        spec = step.generator_spec
        spec.model_kwargs.update({"new_kwarg": 1})
        spec.model_gen_kwargs.update({"new_gen_kwarg": 1})
        self.assertEqual(step.model_kwargs, spec.model_kwargs)
        self.assertEqual(step.model_gen_kwargs, spec.model_gen_kwargs)


class TestGenerationNodeWithBestModelSelector(TestCase):
    def setUp(self) -> None:
        super().setUp()
        self.branin_experiment = get_branin_experiment(
            with_batch=True, with_completed_batch=True
        )
        self.ms_mixed = GeneratorSpec(generator_enum=Generators.BO_MIXED)
        self.ms_botorch = GeneratorSpec(generator_enum=Generators.BOTORCH_MODULAR)

        self.mock_aggregation = MagicMock(
            side_effect=ReductionCriterion.MEAN, spec=ReductionCriterion
        )
        self.model_selection_node = GenerationNode(
            node_name="test",
            generator_specs=[self.ms_mixed, self.ms_botorch],
            best_model_selector=SingleDiagnosticBestModelSelector(
                diagnostic="Fisher exact test p",
                metric_aggregation=self.mock_aggregation,
                criterion=ReductionCriterion.MIN,
            ),
        )

    @mock_botorch_optimize
    def test_gen(self) -> None:
        # Check that with `ModelSelectionNode` generation from a node with
        # multiple model specs does not fail.
        with patch.object(
            self.model_selection_node, "_fit", wraps=self.model_selection_node._fit
        ) as mock_fit:
            gr = self.model_selection_node.gen(
                experiment=self.branin_experiment,
                data=self.branin_experiment.lookup_data(),
                n=1,
                pending_observations={"branin": []},
            )
        # The model specs are practically identical for this example.
        # May pick either one.
        self.assertIsNotNone(gr)
        self.assertEqual(
            self.model_selection_node.model_to_gen_from_name, gr._model_key
        )
        mock_fit.assert_called_with(
            experiment=self.branin_experiment, data=self.branin_experiment.lookup_data()
        )
        # Check that the metric aggregation function is called twice, once for each
        # model spec.
        self.assertEqual(self.mock_aggregation.call_count, 2)

    @mock_botorch_optimize
    def test_pick_fitted_adapter_with_fit_errors(self) -> None:
        # Make model fitting error out for both specs. We should get an error.
        with patch(
            "ax.generation_strategy.generator_spec.GeneratorSpec.fit",
            side_effect=RuntimeError,
        ), self.assertLogs(logger=logger, level="ERROR") as mock_logs:
            self.model_selection_node._fit(experiment=self.branin_experiment)
        self.assertEqual(len(mock_logs.records), 2)
        with self.assertRaisesRegex(ModelError, "No fitted models were found"):
            self.model_selection_node.generator_spec_to_gen_from

        # node._fitted_adapter returns None (rather than erroring out).
        self.assertIsNone(self.model_selection_node._fitted_adapter)

        # Only one spec errors out.
        with patch.object(
            self.ms_mixed, "fit", side_effect=RuntimeError
        ), self.assertLogs(logger=logger, level="ERROR") as mock_logs:
            self.model_selection_node._fit(experiment=self.branin_experiment)
        self.assertEqual(len(mock_logs.records), 1)
        # Picks the model that didn't error out.
        self.assertEqual(
            self.model_selection_node.generator_spec_to_gen_from, self.ms_botorch
        )

    @mock_botorch_optimize
    def test_best_model_selection_errors(self) -> None:
        # Testing that the errors raised within best model selector are
        # gracefully handled. In this case, we'll get an error in CV
        # due to insufficient training data.
        exp = get_branin_experiment(with_completed_trial=True)
        self.model_selection_node._fit(experiment=exp)
        # Check that it selected the first generator and logged a warning.
        with self.assertLogs(logger=logger) as logs:
            self.assertEqual(
                self.model_selection_node.generator_spec_to_gen_from, self.ms_mixed
            )
        self.assertTrue(
            any("raised an error when selecting" in str(log) for log in logs)
        )



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

import numpy as np
import plotly.express as px
from ax.adapter.base import Adapter
from ax.analysis.analysis import Analysis
from ax.analysis.plotly.plotly_analysis import (
    create_plotly_analysis_card,
    PlotlyAnalysisCard,
)
from ax.analysis.plotly.utils import select_metric
from ax.core.experiment import Experiment
from ax.core.map_data import MAP_KEY, MapData
from ax.core.trial_status import TrialStatus
from ax.exceptions.core import UserInputError
from ax.generation_strategy.generation_strategy import GenerationStrategy

from plotly import graph_objects as go
from pyre_extensions import assert_is_instance, override


class ProgressionPlot(Analysis):
    """
    Plotly Scatter showing a timerseries-like metric's progression, with one line for
    each arm. The plot also includes a marker on the terminal step of any trial that
    was early stopped.

    The DataFrame computed will contain one row per arm and the following columns:
        - trial_index: The trial index of the arm
        - arm_name: The name of the arm
        - METRIC_NAME: The observed mean of the metric specified
        - progression: The progression at which the metric was observed
        - wallclock_time: The wallclock time at which the metric was observed, in
            seconds and starting at 0 from the first trial's start time.
    """

    def __init__(
        self, metric_name: str | None = None, by_wallclock_time: bool = False
    ) -> None:
        """
        Args:
            metric_name: The name of the metric to plot. If not specified the objective
                will be used. Note that the metric cannot be inferred for
                multi-objective or scalarized-objective experiments.
            wallclock_time: If True, plot the relative wallclock time instead of the
                progression on the x-axis.
        """

        self._metric_name = metric_name
        self._by_wallclock_time = by_wallclock_time

    @override
    def compute(
        self,
        experiment: Experiment | None = None,
        generation_strategy: GenerationStrategy | None = None,
        adapter: Adapter | None = None,
    ) -> PlotlyAnalysisCard:
        if experiment is None:
            raise UserInputError("ProgressionPlot requires an Experiment")

        data = experiment.lookup_data()
        if not isinstance(data, MapData):
            raise UserInputError("ProgressionPlot requires MapData")

        metric_name = self._metric_name or select_metric(experiment=experiment)

        # Collect the data necessary to plot each progression curve.
        map_df = data.map_df
        df = map_df.loc[
            map_df["metric_name"] == metric_name,
            ["trial_index", "arm_name", "mean", MAP_KEY],
        ].rename(columns={MAP_KEY: "progression", "mean": metric_name})

        # Get the terminal step of each trial that was early stopped so we can place a
        # marker to inform the user.
        data_df = data.df  # Collect dataframe with only the terminal observations
        terminal_points = data_df.loc[
            data_df["trial_index"].isin(
                [
                    trial.index
                    for trial in experiment.trials_by_status[TrialStatus.EARLY_STOPPED]
                ]
            ),
            ["trial_index", "mean", MAP_KEY],
        ].rename(columns={MAP_KEY: "progression", "mean": metric_name})

        # Add the wallclock time column
        wallclock_series = _calculate_wallclock_timeseries(
            experiment=experiment, metric_name=metric_name
        )

        # If there is a nan in the wallclock time dict's keys then lookup will fail. If
        # this happens, set the wallclock time to nan and continue.
        df["wallclock_time"] = df.apply(
            lambda row: wallclock_series[row["trial_index"]].get(
                row["progression"], np.nan
            ),
            axis=1,
        )
        if len(terminal_points) > 0:
            terminal_points["wallclock_time"] = terminal_points.apply(
                lambda row: wallclock_series[row["trial_index"]].get(
                    row["progression"], np.nan
                ),
                axis=1,
            )

        # Plot the progression lines with one curve for each arm.
        if self._by_wallclock_time:
            x_axis_name = "wallclock_time"
        else:
            x_axis_name = "progression"

        fig = px.line(df, x=x_axis_name, y=metric_name, color="arm_name")

        # Add a marker for each terminal point on early stopped trials.
        if len(terminal_points) > 0:
            fig.add_trace(
                go.Scatter(
                    x=terminal_points[x_axis_name],
                    y=terminal_points[metric_name],
                    mode="markers",
                    marker={"symbol": "x", "color": "red"},
                    showlegend=False,
                    hoverinfo="none",
                )
            )

        return create_plotly_analysis_card(
            name=self.__class__.__name__,
            title=f"{metric_name} by {x_axis_name.replace('_', ' ')}",
            subtitle=(
                "The progression plot tracks the evolution of each metric "
                "over the course of the experiment. This visualization is "
                "typically used to monitor the improvement of metrics over "
                "Trial iterations, but can also be useful in informing decisions "
                "about early stopping for Trials."
            ),
            df=df,
            fig=fig,
        )


def _calculate_wallclock_timeseries(
    experiment: Experiment,
    metric_name: str,
) -> dict[int, dict[float, float]]:
    """
    Calculate a mapping from each trial index and progression to the time since the
    first trial started, in seconds. Assume that the first trial started at t=0, and
    that progressions are linearly spaced between the start and completion times of
    each trial.

    If a trial does not have either a start or completion time the wallclock time
    cannot be calculated and the value will be nan (which will not be plotted).

    Returns:
        trial_index => (progression => timestamp)
    """
    # Find the earliest start time.
    start_time = min(
        trial.time_run_started.timestamp()
        for trial in experiment.trials.values()
        if trial.time_run_started is not None
    )
    # Calculate all start and completion times relative to the earliest start time.
    # Give nan for trials that don't have a start or completion time.
    relative_timestamps = {
        idx: (
            trial.time_run_started.timestamp() - start_time
            if trial.time_run_started is not None
            else np.nan,
            trial.time_completed.timestamp() - start_time
            if trial.time_completed is not None
            else np.nan,
        )
        for idx, trial in experiment.trials.items()
    }

    data = assert_is_instance(experiment.lookup_data(), MapData)
    df = data.map_df[data.map_df["metric_name"] == metric_name]

    return {
        trial_index: dict(
            zip(
                df[df["trial_index"] == trial_index][MAP_KEY].to_numpy(),
                # Map the progressions to linspace if the start and completion times
                # are both available, otherwise map to nans
                np.linspace(
                    relative_timestamps[trial_index][0],
                    relative_timestamps[trial_index][1],
                    len(df[df["trial_index"] == trial_index]),
                )
                if (
                    relative_timestamps[trial_index][0] is not None
                    and relative_timestamps[trial_index][1] is not None
                )
                else np.full(len(df[df["trial_index"] == trial_index]), np.nan),
            )
        )
        for trial_index in experiment.trials.keys()
    }
